from collections import defaultdict
from urllib.parse import urlparse

import code
import functools
import json
import re
import os
import platform
import pdb
import shutil
import sys
import time
import traceback
import runpy
import webbrowser
import contextlib
from pathlib import Path
from datetime import datetime

import click
import github3
import pkg_resources
import requests
from requests import exceptions
from rst2ansi import rst2ansi
from jinja2 import Environment
from jinja2 import PackageLoader

import cumulusci
from cumulusci.core.config import OrgConfig
from cumulusci.core.config import ScratchOrgConfig
from cumulusci.core.config import ServiceConfig
from cumulusci.core.config import TaskConfig
from cumulusci.core.config import UniversalConfig
from cumulusci.core.github import create_gist, get_github_api
from cumulusci.core.exceptions import OrgNotFound
from cumulusci.core.exceptions import CumulusCIException
from cumulusci.core.exceptions import CumulusCIUsageError
from cumulusci.core.exceptions import ServiceNotConfigured
from cumulusci.core.exceptions import FlowNotFoundError


from cumulusci.core.utils import import_global
from cumulusci.cli.runtime import CliRuntime
from cumulusci.cli.runtime import get_installed_version
from cumulusci.cli.ui import CliTable, CROSSMARK, SimpleSalesforceUIHelpers
from cumulusci.salesforce_api.utils import get_simple_salesforce_connection
from cumulusci.utils import doc_task
from cumulusci.utils import parse_api_datetime
from cumulusci.utils import get_cci_upgrade_command
from cumulusci.utils.git import current_branch
from cumulusci.utils.logging import tee_stdout_stderr
from cumulusci.oauth.salesforce import CaptureSalesforceOAuth
from cumulusci.core.utils import cleanup_org_cache_dirs


from .logger import init_logger, get_tempfile_logger


@contextlib.contextmanager
def timestamp_file():
    """Opens a file for tracking the time of the last version check"""

    config_dir = UniversalConfig.default_cumulusci_dir()
    timestamp_file = os.path.join(config_dir, "cumulus_timestamp")

    try:
        with open(timestamp_file, "r+") as f:
            yield f
    except IOError:  # file does not exist
        with open(timestamp_file, "w+") as f:
            yield f


FINAL_VERSION_RE = re.compile(r"^[\d\.]+$")


def is_final_release(version: str) -> bool:
    """Returns bool whether version string should be considered a final release.

    cumulusci versions are considered final if they contain only digits and periods.
    e.g. 1.0.1 is final but 2.0b1 and 2.0.dev0 are not.
    """
    return bool(FINAL_VERSION_RE.match(version))


def get_latest_final_version():
    """ return the latest version of cumulusci in pypi, be defensive """
    # use the pypi json api https://wiki.python.org/moin/PyPIJSON
    res = requests.get("https://pypi.org/pypi/cumulusci/json", timeout=5).json()
    with timestamp_file() as f:
        f.write(str(time.time()))
    versions = []
    for versionstring in res["releases"].keys():
        if not is_final_release(versionstring):
            continue
        versions.append(pkg_resources.parse_version(versionstring))
    versions.sort(reverse=True)
    return versions[0]


def check_latest_version():
    """ checks for the latest version of cumulusci from pypi, max once per hour """
    check = True

    with timestamp_file() as f:
        timestamp = float(f.read() or 0)
    delta = time.time() - timestamp
    check = delta > 3600

    if check:
        try:
            latest_version = get_latest_final_version()
        except requests.exceptions.RequestException as e:
            click.echo("Error checking cci version:")
            click.echo(str(e))
            return

        result = latest_version > get_installed_version()
        click.echo("Checking the version!")
        if result:
            click.echo(
                f"""An update to CumulusCI is available. To install the update, run this command: {get_cci_upgrade_command()}"""
            )


def render_recursive(data, indent=None):
    if indent is None:
        indent = 0
    indent_str = " " * indent
    if isinstance(data, list):
        for item in data:
            if isinstance(item, (bytes, str)):
                click.echo(f"{indent_str}- {item}")
            else:
                click.echo(f"{indent_str}-")
                render_recursive(item, indent=indent + 4)
    elif isinstance(data, dict):
        for key, value in data.items():
            key_str = click.style(str(key) + ":", bold=True)
            if isinstance(value, list):
                click.echo(f"{indent_str}{key_str}")
                render_recursive(value, indent=indent + 4)
            elif isinstance(value, dict):
                click.echo(f"{indent_str}{key_str}")
                render_recursive(value, indent=indent + 4)
            else:
                click.echo(f"{indent_str}{key_str} {value}")


# global reference to active runtime
RUNTIME = None


def pass_runtime(func=None, require_project=True, require_keychain=False):
    """Decorator which passes the CCI runtime object as the first arg to a click command."""

    def decorate(func):
        def new_func(*args, **kw):
            runtime = RUNTIME
            if require_project and runtime.project_config is None:
                raise runtime.project_config_error
            if require_keychain:
                runtime._load_keychain()
            func(RUNTIME, *args, **kw)

        return functools.update_wrapper(new_func, func)

    if func is None:
        return decorate
    else:
        return decorate(func)


SUGGEST_ERROR_COMMAND = (
    """Run this command for more information about debugging errors: cci error --help"""
)

USAGE_ERRORS = (CumulusCIUsageError, click.UsageError)


#
# Root command
#
def main(args=None):
    """Main CumulusCI CLI entry point.

    This runs as the first step in processing any CLI command.

    This wraps the `click` library in order to do some initialization and centralized error handling.
    """
    with contextlib.ExitStack() as stack:
        args = args or sys.argv
        # Check for updates _unless_ we've been asked to output JSON,
        # or if we're going to check anyway as part of the `version` command.
        is_version_command = len(args) > 1 and args[1] == "version"
        if "--json" not in args and not is_version_command:
            check_latest_version()

        # Load CCI config
        global RUNTIME
        RUNTIME = CliRuntime(load_keychain=False)
        RUNTIME.check_cumulusci_version()

        # Configure logging
        debug = "--debug" in args
        if debug:
            args.remove("--debug")
        should_show_stacktraces = RUNTIME.universal_config.cli__show_stacktraces

        # Only create logfiles for commands
        # that are not `cci error`
        is_error_command = len(args) > 2 and args[1] == "error"
        tempfile_path = None
        if not is_error_command:
            logger, tempfile_path = get_tempfile_logger()
            stack.enter_context(tee_stdout_stderr(args, logger, tempfile_path))

        init_logger(log_requests=debug)
        # Hand CLI processing over to click, but handle exceptions
        try:
            cli(standalone_mode=False)
        except click.Abort:  # Keyboard interrupt
            show_debug_info() if debug else click.echo("\nAborted!")
            sys.exit(1)
        except Exception as e:
            if debug:
                show_debug_info()
            else:
                handle_exception(
                    e, is_error_command, tempfile_path, should_show_stacktraces
                )
            sys.exit(1)


def handle_exception(error, is_error_cmd, logfile_path, should_show_stacktraces=False):
    """Displays error of appropriate message back to user, prompts user to investigate further
    with `cci error` commands, and writes the traceback to the latest logfile.
    """
    if isinstance(error, exceptions.ConnectionError):
        connection_error_message()
    elif isinstance(error, click.ClickException):
        click.echo(click.style(f"Error: {error.format_message()}", fg="red"))
    else:
        click.echo(click.style(f"Error: {error}", fg="red"))
    # Only suggest gist command if it wasn't run
    if not is_error_cmd:
        click.echo(click.style(SUGGEST_ERROR_COMMAND, fg="yellow"))

    # This is None if we're handling an exception for a
    # `cci error` command.
    if logfile_path:
        with open(logfile_path, "a") as log_file:
            traceback.print_exc(file=log_file)  # log stacktrace silently

    if should_show_stacktraces and not isinstance(error, USAGE_ERRORS):
        raise error


def connection_error_message():
    message = (
        "We encountered an error with your internet connection. "
        "Please check your connection and try the last cci command again."
    )
    click.echo(click.style(message, fg="red"))


def show_debug_info():
    """Displays the traceback and opens pdb"""
    traceback.print_exc()
    pdb.post_mortem()


@click.group("main", help="")
def cli():
    """Top-level `click` command group."""


@cli.command(name="version", help="Print the current version of CumulusCI")
def version():
    click.echo("CumulusCI version: ", nl=False)
    click.echo(click.style(cumulusci.__version__, bold=True), nl=False)
    click.echo(f" ({sys.argv[0]})")
    click.echo(f"Python version: {sys.version.split()[0]}", nl=False)
    click.echo(f" ({sys.executable})")

    click.echo()
    current_version = get_installed_version()
    latest_version = get_latest_final_version()
    if latest_version > current_version:
        click.echo(
            f"There is a newer version of CumulusCI available ({str(latest_version)})."
        )
        click.echo(f"To upgrade, run `{get_cci_upgrade_command()}`")
        click.echo(
            f"Release notes: https://github.com/SFDO-Tooling/CumulusCI/releases/tag/v{str(latest_version)}"
        )
    else:
        click.echo("You have the latest version of CumulusCI.")

    click.echo()


@cli.command(name="shell", help="Drop into a Python shell")
@click.option("--script", help="Path to a script to run", type=click.Path())
@click.option("--python", help="Python code to run directly")
@pass_runtime(require_project=False, require_keychain=True)
def shell(runtime, script=None, python=None):
    # alias for backwards-compatibility
    variables = {
        "config": runtime,
        "runtime": runtime,
        "project_config": runtime.project_config,
    }

    if script:
        if python:
            raise click.UsageError("Cannot specify both --script and --python")
        runpy.run_path(script, init_globals=variables)
    elif python:
        exec(python, variables)
    else:
        code.interact(local=variables)


GIST_404_ERR_MSG = """A 404 error code was returned when trying to create your gist.
Please ensure that your GitHub personal access token has the 'Create gists' scope."""


def get_context_info():
    host_info = platform.uname()

    info = []
    info.append(f"CumulusCI version: {cumulusci.__version__}")
    info.append(f"Python version: {sys.version.split()[0]} ({sys.executable})")
    info.append(f"Environment Info: {host_info.system} / {host_info.machine}")
    return "\n".join(info)


# Top Level Groups


@cli.group(
    "project", help="Commands for interacting with project repository configurations"
)
def project():
    pass


@cli.group("org", help="Commands for connecting and interacting with Salesforce orgs")
def org():
    pass


@cli.group("task", help="Commands for finding and running tasks for a project")
def task():
    pass


@cli.group("flow", help="Commands for finding and running flows for a project")
def flow():
    pass


@cli.group("service", help="Commands for connecting services to the keychain")
def service():
    pass


@cli.group("error", short_help="Get or share information about an error")
def error():
    """
    Get or share information about an error

    If you'd like to dig into an error more yourself,
    you can get the last few lines of context about it
    from `cci error info`.

    If you'd like to submit it to a developer for conversation,
    you can use the `cci error gist` command. Just make sure
    that your GitHub access token has the 'create gist' scope.

    If you'd like to regularly see stack traces, set the `show_stacktraces`
    option to `True` in the "cli" section of `~/.cumulusci/cumulusci.yml`, or to
    see a stack-trace (and other debugging information) just once, use the `--debug`
    command line option.

    For more information on working with errors in CumulusCI visit:
    https://cumulusci.readthedocs.io/en/latest/features.html#working-with-errors
    """


# Commands for group: project


def validate_project_name(value):
    if not re.match(r"^[a-zA-Z0-9_-]+$", value):
        raise click.UsageError(
            "Invalid project name. Allowed characters: "
            "letters, numbers, dash, and underscore"
        )
    return value


@project.command(
    name="init", help="Initialize a new project for use with the cumulusci toolbelt"
)
@pass_runtime(require_project=False)
def project_init(runtime):
    if not os.path.isdir(".git"):
        raise click.ClickException("You are not in the root of a Git repository")

    if os.path.isfile("cumulusci.yml"):
        raise click.ClickException("This project already has a cumulusci.yml file")

    context = {"cci_version": cumulusci.__version__}

    # Prep jinja2 environment for rendering files
    env = Environment(
        loader=PackageLoader(
            "cumulusci", os.path.join("files", "templates", "project")
        ),
        trim_blocks=True,
        lstrip_blocks=True,
    )

    # Project and Package Info
    click.echo()
    click.echo(click.style("# Project Info", bold=True, fg="blue"))
    click.echo(
        "The following prompts will collect general information about the project"
    )

    project_name = os.path.split(os.getcwd())[-1:][0]
    click.echo()
    click.echo(
        "Enter the project name.  The name is usually the same as your repository name.  NOTE: Do not use spaces in the project name!"
    )
    context["project_name"] = click.prompt(
        click.style("Project Name", bold=True),
        default=project_name,
        value_proc=validate_project_name,
    )

    click.echo()
    click.echo(
        "CumulusCI uses an unmanaged package as a container for your project's metadata.  Enter the name of the package you want to use."
    )
    context["package_name"] = click.prompt(
        click.style("Package Name", bold=True), default=project_name
    )

    click.echo()
    context["package_namespace"] = None
    if click.confirm(
        click.style("Is this a managed package project?", bold=True), default=False
    ):
        click.echo(
            "Enter the namespace assigned to the managed package for this project"
        )
        context["package_namespace"] = click.prompt(
            click.style("Package Namespace", bold=True), default=project_name
        )

    click.echo()
    context["api_version"] = click.prompt(
        click.style("Salesforce API Version", bold=True),
        default=runtime.universal_config.project__package__api_version,
    )

    click.echo()
    click.echo(
        "Salesforce metadata can be stored using Metadata API format or DX source format. "
        "Which do you want to use?"
    )
    context["source_format"] = click.prompt(
        click.style("Source format", bold=True),
        type=click.Choice(["sfdx", "mdapi"]),
        default="sfdx",
    )

    # Dependencies
    dependencies = []
    click.echo(click.style("# Extend Project", bold=True, fg="blue"))
    click.echo(
        "CumulusCI makes it easy to build extensions of other projects configured for CumulusCI like Salesforce.org's NPSP and EDA.  If you are building an extension of another project using CumulusCI and have access to its Github repository, use this section to configure this project as an extension."
    )
    if click.confirm(
        click.style(
            "Are you extending another CumulusCI project such as NPSP or EDA?",
            bold=True,
        ),
        default=False,
    ):
        click.echo("Please select from the following options:")
        click.echo("  1: EDA (https://github.com/SalesforceFoundation/EDA)")
        click.echo("  2: NPSP (https://github.com/SalesforceFoundation/NPSP)")
        click.echo(
            "  3: Github URL (provide a URL to a Github repository configured for CumulusCI)"
        )
        selection = click.prompt(click.style("Enter your selection", bold=True))
        github_url = {
            "1": "https://github.com/SalesforceFoundation/EDA",
            "2": "https://github.com/SalesforceFoundation/NPSP",
        }.get(selection)
        if github_url is None:
            print(selection)
            github_url = click.prompt(
                click.style("Enter the Github Repository URL", bold=True)
            )
        dependencies.append({"type": "github", "url": github_url})
    context["dependencies"] = dependencies

    # Git Configuration
    git_config = {}
    click.echo()
    click.echo(click.style("# Git Configuration", bold=True, fg="blue"))
    click.echo(
        "CumulusCI assumes the current git branch is your default branch, your feature branches are named feature/*, your beta release tags are named beta/*, and your release tags are release/*.  If you want to use a different branch/tag naming scheme, you can configure the overrides here.  Otherwise, just accept the defaults."
    )

    default_main_branch = current_branch(os.getcwd()) or "main"
    git_default_branch = click.prompt(
        click.style("Default Branch", bold=True), default=default_main_branch
    )
    if (
        git_default_branch
        and git_default_branch != runtime.universal_config.project__git__default_branch
    ):
        git_config["default_branch"] = git_default_branch

    git_prefix_feature = click.prompt(
        click.style("Feature Branch Prefix", bold=True), default="feature/"
    )
    if (
        git_prefix_feature
        and git_prefix_feature != runtime.universal_config.project__git__prefix_feature
    ):
        git_config["prefix_feature"] = git_prefix_feature

    git_prefix_beta = click.prompt(
        click.style("Beta Tag Prefix", bold=True), default="beta/"
    )
    if (
        git_prefix_beta
        and git_prefix_beta != runtime.universal_config.project__git__prefix_beta
    ):
        git_config["prefix_beta"] = git_prefix_beta

    git_prefix_release = click.prompt(
        click.style("Release Tag Prefix", bold=True), default="release/"
    )
    if (
        git_prefix_release
        and git_prefix_release != runtime.universal_config.project__git__prefix_release
    ):
        git_config["prefix_release"] = git_prefix_release

    context["git"] = git_config

    #     test:
    click.echo()
    click.echo(click.style("# Apex Tests Configuration", bold=True, fg="blue"))
    click.echo(
        "The CumulusCI Apex test runner uses a SOQL where clause to select which tests to run.  Enter the SOQL pattern to use to match test class names."
    )

    test_name_match = click.prompt(
        click.style("Test Name Match", bold=True),
        default=runtime.universal_config.project__test__name_match,
    )
    if (
        test_name_match
        and test_name_match == runtime.universal_config.project__test__name_match
    ):
        test_name_match = None
    context["test_name_match"] = test_name_match

    context["code_coverage"] = None
    if click.confirm(
        click.style(
            "Do you want to check Apex code coverage when tests are run?", bold=True
        ),
        default=True,
    ):
        context["code_coverage"] = click.prompt(
            click.style("Minimum code coverage percentage", bold=True), default=75
        )

    # Render templates
    for name in (".gitignore", "README.md", "cumulusci.yml"):
        template = env.get_template(name)
        with open(name, "w") as f:
            f.write(template.render(**context))

    # Create source directory
    source_path = "force-app" if context["source_format"] == "sfdx" else "src"
    if not os.path.isdir(source_path):
        os.mkdir(source_path)

    # Create sfdx-project.json
    if not os.path.isfile("sfdx-project.json"):

        sfdx_project = {
            "packageDirectories": [{"path": "force-app", "default": True}],
            "namespace": context["package_namespace"],
            "sourceApiVersion": context["api_version"],
        }
        with open("sfdx-project.json", "w") as f:
            f.write(json.dumps(sfdx_project))

    # Create orgs subdir
    if not os.path.isdir("orgs"):
        os.mkdir("orgs")

    template = env.get_template("scratch_def.json")
    with open(os.path.join("orgs", "beta.json"), "w") as f:
        f.write(
            template.render(
                package_name=context["package_name"],
                org_name="Beta Test Org",
                edition="Developer",
                managed=True,
            )
        )
    with open(os.path.join("orgs", "dev.json"), "w") as f:
        f.write(
            template.render(
                package_name=context["package_name"],
                org_name="Dev Org",
                edition="Developer",
                managed=False,
            )
        )
    with open(os.path.join("orgs", "feature.json"), "w") as f:
        f.write(
            template.render(
                package_name=context["package_name"],
                org_name="Feature Test Org",
                edition="Developer",
                managed=False,
            )
        )
    with open(os.path.join("orgs", "release.json"), "w") as f:
        f.write(
            template.render(
                package_name=context["package_name"],
                org_name="Release Test Org",
                edition="Enterprise",
                managed=True,
            )
        )

    # create robot folder structure and starter files
    if not os.path.isdir("robot"):
        test_folder = os.path.join("robot", context["project_name"], "tests")
        resource_folder = os.path.join("robot", context["project_name"], "resources")
        doc_folder = os.path.join("robot", context["project_name"], "doc")

        os.makedirs(test_folder)
        os.makedirs(resource_folder)
        os.makedirs(doc_folder)
        test_src = os.path.join(
            cumulusci.__location__,
            "robotframework",
            "tests",
            "salesforce",
            "create_contact.robot",
        )
        test_dest = os.path.join(test_folder, "create_contact.robot")
        shutil.copyfile(test_src, test_dest)

    # Create pull request template
    if not os.path.isdir(".github"):
        os.mkdir(".github")
        with open(os.path.join(".github", "PULL_REQUEST_TEMPLATE.md"), "w") as f:
            f.write(
                """

# Critical Changes

# Changes

# Issues Closed
"""
            )

    # Create datasets folder
    if not os.path.isdir("datasets"):
        os.mkdir("datasets")
        template = env.get_template("mapping.yml")
        with open(os.path.join("datasets", "mapping.yml"), "w") as f:
            f.write(template.render(**context))

    click.echo(
        click.style(
            "Your project is now initialized for use with CumulusCI",
            bold=True,
            fg="green",
        )
    )


@project.command(
    name="info", help="Display information about the current project's configuration"
)
@pass_runtime
def project_info(runtime):
    render_recursive(runtime.project_config.project)


@project.command(
    name="dependencies",
    help="Displays the current dependencies for the project.  If the dependencies section has references to other github repositories, the repositories are inspected and a static list of dependencies is created",
)
@pass_runtime(require_keychain=True)
def project_dependencies(runtime):
    dependencies = runtime.project_config.get_static_dependencies()
    for line in runtime.project_config.pretty_dependencies(dependencies):
        click.echo(line)


# Commands for group: service


@service.command(name="list", help="List services available for configuration and use")
@click.option("--plain", is_flag=True, help="Print the table using plain ascii.")
@click.option("--json", "print_json", is_flag=True, help="Print a json string")
@pass_runtime(require_project=False, require_keychain=True)
def service_list(runtime, plain, print_json):
    services = (
        runtime.project_config.services
        if runtime.project_config is not None
        else runtime.universal_config.services
    )
    configured_services = runtime.keychain.list_services()
    plain = plain or runtime.universal_config.cli__plain_output

    data = [["Name", "Description", "Configured"]]
    for serv, schema in services.items():
        schema["configured"] = serv in configured_services
        data.append([serv, schema["description"], schema["configured"]])

    if print_json:
        click.echo(json.dumps(services))
        return None

    rows_to_dim = [row_index for row_index, row in enumerate(data) if not row[2]]
    table = CliTable(
        data,
        title="Services",
        wrap_cols=["Description"],
        bool_cols=["Configured"],
        dim_rows=rows_to_dim,
    )
    table.echo(plain)


class ConnectServiceCommand(click.MultiCommand):
    def _get_services_config(self, runtime):
        return (
            runtime.project_config.services
            if runtime.project_config
            else runtime.universal_config.services
        )

    def list_commands(self, ctx):
        """ list the services that can be configured """
        services = self._get_services_config(RUNTIME)
        return sorted(services.keys())

    def _build_param(self, attribute, details):
        req = details["required"]
        return click.Option((f"--{attribute}",), prompt=req, required=req)

    def get_command(self, ctx, name):
        runtime = RUNTIME
        runtime._load_keychain()
        services = self._get_services_config(RUNTIME)
        try:
            service_config = services[name]
        except KeyError:
            raise click.UsageError(f"Sorry, I don't know about the '{name}' service.")
        attributes = service_config["attributes"].items()

        params = [self._build_param(attr, cnfg) for attr, cnfg in attributes]
        if runtime.project_config is not None:
            params.append(click.Option(("--project",), is_flag=True))

        def callback(*args, **kwargs):
            if runtime.project_config is None:
                project = False
            else:
                project = kwargs.pop("project", False)
            serv_conf = dict(
                (k, v) for k, v in list(kwargs.items()) if v is not None
            )  # remove None values

            # A service can define a callable to validate the service config
            validator_path = service_config.get("validator")
            if validator_path:
                validator = import_global(validator_path)
                validator(serv_conf)

            runtime.keychain.set_service(name, ServiceConfig(serv_conf), project)
            if project:
                click.echo(f"{name} is now configured for this project.")
            else:
                click.echo(f"{name} is now configured for all CumulusCI projects.")

        ret = click.Command(name, params=params, callback=callback)
        return ret


@service.command(
    cls=ConnectServiceCommand, name="connect", help="Connect a CumulusCI task service"
)
def service_connect():
    pass


@service.command(name="info", help="Show the details of a connected service")
@click.argument("service_name")
@click.option("--plain", is_flag=True, help="Print the table using plain ascii.")
@pass_runtime(require_project=False, require_keychain=True)
def service_info(runtime, service_name, plain):
    try:
        plain = plain or runtime.universal_config.cli__plain_output
        service_config = runtime.keychain.get_service(service_name)
        service_data = [["Key", "Value"]]
        service_data.extend(
            [
                [click.style(k, bold=True), str(v)]
                for k, v in service_config.config.items()
            ]
        )
        wrap_cols = ["Value"] if not plain else None
        service_table = CliTable(service_data, title=service_name, wrap_cols=wrap_cols)
        service_table._table.inner_heading_row_border = False
        service_table.echo(plain)
    except ServiceNotConfigured:
        click.echo(
            "{0} is not configured for this project.  Use service connect {0} to configure.".format(
                service_name
            )
        )


# Commands for group: org


@org.command(
    name="browser",
    help="Opens a browser window and logs into the org using the stored OAuth credentials",
)
@click.argument("org_name", required=False)
@pass_runtime(require_project=False, require_keychain=True)
def org_browser(runtime, org_name):
    org_name, org_config = runtime.get_org(org_name)
    org_config.refresh_oauth_token(runtime.keychain)

    webbrowser.open(org_config.start_url)
    # Save the org config in case it was modified
    org_config.save()


@org.command(
    name="connect", help="Connects a new org's credentials using OAuth Web Flow"
)
@click.argument("org_name")
@click.option(
    "--sandbox", is_flag=True, help="If set, connects to a Salesforce sandbox org"
)
@click.option(
    "--login-url",
    help="If set, login to this hostname.",
    default="https://login.salesforce.com",
)
@click.option(
    "--default",
    is_flag=True,
    help="If set, sets the connected org as the new default org",
)
@click.option(
    "--global-org", help="Set True if org should be used by any project", is_flag=True
)
@pass_runtime(require_project=False, require_keychain=True)
def org_connect(runtime, org_name, sandbox, login_url, default, global_org):
    runtime.check_org_overwrite(org_name)

    if "lightning.force.com" in login_url:
        raise click.UsageError(
            "Connecting an org with a lightning.force.com URL does not work. "
            "Use the my.salesforce.com version instead"
        )

    connected_app = runtime.keychain.get_service("connected_app")
    if sandbox:
        login_url = "https://test.salesforce.com"

    oauth_capture = CaptureSalesforceOAuth(
        client_id=connected_app.client_id,
        client_secret=connected_app.client_secret,
        callback_url=connected_app.callback_url,
        auth_site=login_url,
        scope="web full refresh_token",
    )
    oauth_dict = oauth_capture()
    global_org = global_org or runtime.project_config is None
    org_config = OrgConfig(oauth_dict, org_name, runtime.keychain, global_org)
    org_config.load_userinfo()
    org_config._load_orginfo()
    if org_config.organization_sobject["TrialExpirationDate"] is None:
        org_config.config["expires"] = "Persistent"
    else:
        org_config.config["expires"] = parse_api_datetime(
            org_config.organization_sobject["TrialExpirationDate"]
        ).date()

    org_config.save()

    if default and runtime.project_config is not None:
        runtime.keychain.set_default_org(org_name)
        click.echo(f"{org_name} is now the default org")


@org.command(name="default", help="Sets an org as the default org for tasks and flows")
@click.argument("org_name")
@click.option(
    "--unset",
    is_flag=True,
    help="Unset the org as the default org leaving no default org selected",
)
@pass_runtime(require_keychain=True)
def org_default(runtime, org_name, unset):
    if unset:
        runtime.keychain.unset_default_org()
        click.echo(f"{org_name} is no longer the default org.  No default org set.")
    else:
        runtime.keychain.set_default_org(org_name)
        click.echo(f"{org_name} is now the default org")


@org.command(name="import", help="Import a scratch org from Salesforce DX")
@click.argument("username_or_alias")
@click.argument("org_name")
@pass_runtime(require_keychain=True)
def org_import(runtime, username_or_alias, org_name):
    org_config = {"username": username_or_alias}
    scratch_org_config = ScratchOrgConfig(
        org_config, org_name, runtime.keychain, global_org=False
    )
    scratch_org_config.config["created"] = True

    info = scratch_org_config.sfdx_info
    scratch_org_config.config["days"] = calculate_org_days(info)
    scratch_org_config.config["date_created"] = parse_api_datetime(info["created_date"])

    scratch_org_config.save()
    click.echo(
        "Imported scratch org: {org_id}, username: {username}".format(
            **scratch_org_config.sfdx_info
        )
    )


def calculate_org_days(info):
    """Returns the difference in days between created_date (ISO 8601),
    and expiration_date (%Y-%m-%d)"""
    if not info.get("created_date") or not info.get("expiration_date"):
        return 1
    created_date = parse_api_datetime(info["created_date"]).date()
    expires_date = datetime.strptime(info["expiration_date"], "%Y-%m-%d").date()
    return abs((expires_date - created_date).days)


@org.command(name="info", help="Display information for a connected org")
@click.argument("org_name", required=False)
@click.option(
    "print_json", "--json", is_flag=True, help="Print as JSON.  Includes access token"
)
@pass_runtime(require_project=False, require_keychain=True)
def org_info(runtime, org_name, print_json):
    org_name, org_config = runtime.get_org(org_name)
    org_config.refresh_oauth_token(runtime.keychain)

    if print_json:
        click.echo(
            json.dumps(
                org_config.config,
                sort_keys=True,
                indent=4,
                default=str,
                separators=(",", ": "),
            )
        )
    else:
        UI_KEYS = [
            "config_file",
            "config_name",
            "created",
            "date_created",
            "days",
            "default",
            "email_address",
            "instance_url",
            "instance_name",
            "is_sandbox",
            "namespaced",
            "org_id",
            "org_type",
            "password",
            "scratch",
            "scratch_org_type",
            "set_password",
            "sfdx_alias",
            "username",
        ]
        keys = [key for key in org_config.config.keys() if key in UI_KEYS]
        pairs = [[key, str(org_config.config[key])] for key in keys]
        pairs.append(["api_version", org_config.latest_api_version])
        pairs.sort()
        table_data = [["Key", "Value"]]
        table_data.extend(
            [[click.style(key, bold=True), value] for key, value in pairs]
        )
        table = CliTable(table_data, wrap_cols=["Value"])
        table.echo()

        if org_config.scratch and org_config.expires:
            click.echo("Org expires on {:%c}".format(org_config.expires))

    # Save the org config in case it was modified
    org_config.save()


@org.command(name="list", help="Lists the connected orgs for the current project")
@click.option("--plain", is_flag=True, help="Print the table using plain ascii.")
@pass_runtime(require_project=False, require_keychain=True)
def org_list(runtime, plain):
    plain = plain or runtime.universal_config.cli__plain_output
    header = ["Name", "Default", "Username", "Expires"]
    persistent_data = [header]
    scratch_data = [header[:2] + ["Days", "Expired", "Config", "Domain"]]
    org_configs = {
        org: runtime.keychain.get_org(org) for org in runtime.keychain.list_orgs()
    }
    rows_to_dim = []
    default_org_name, _ = runtime.keychain.get_default_org()
    for org, org_config in org_configs.items():
        row = [org, org == default_org_name]
        if isinstance(org_config, ScratchOrgConfig):
            org_days = org_config.format_org_days()
            if org_config.expired:
                domain = ""
            else:
                instance_url = org_config.config.get("instance_url", "")
                domain = urlparse(instance_url).hostname or ""
                if domain:
                    domain = domain.replace(".my.salesforce.com", "")
            row.extend(
                [org_days, not org_config.active, org_config.config_name, domain]
            )
            scratch_data.append(row)
        else:
            username = org_config.config.get(
                "username", org_config.userinfo__preferred_username
            )
            row.append(username)
            row.append(org_config.expires or "Unknown")
            persistent_data.append(row)

    rows_to_dim = [row_index for row_index, row in enumerate(scratch_data) if row[3]]
    scratch_table = CliTable(
        scratch_data, title="Scratch Orgs", bool_cols=["Default"], dim_rows=rows_to_dim
    )
    scratch_table.stringify_boolean_col(col_name="Expired", true_str=CROSSMARK)
    scratch_table.echo(plain)

    wrap_cols = ["Username"] if not plain else None
    persistent_table = CliTable(
        persistent_data,
        title="Connected Orgs",
        wrap_cols=wrap_cols,
        bool_cols=["Default"],
    )
    persistent_table.echo(plain)
    cleanup_org_cache_dirs(runtime.keychain, runtime.project_config)


@org.command(
    name="prune", help="Removes all expired scratch orgs from the current project"
)
@click.option(
    "--include-active",
    is_flag=True,
    help="Remove all scratch orgs, regardless of expiry.",
)
@pass_runtime(require_project=True, require_keychain=True)
def org_prune(runtime, include_active=False):

    predefined_scratch_configs = getattr(runtime.project_config, "orgs__scratch", {})

    expired_orgs_removed = []
    active_orgs_removed = []
    org_shapes_skipped = []
    active_orgs_skipped = []
    for org_name in runtime.keychain.list_orgs():

        org_config = runtime.keychain.get_org(org_name)

        if org_name in predefined_scratch_configs:
            if org_config.active and include_active:
                runtime.keychain.remove_org(org_name)
                active_orgs_removed.append(org_name)
            else:
                org_shapes_skipped.append(org_name)

        elif org_config.active:
            if include_active:
                runtime.keychain.remove_org(org_name)
                active_orgs_removed.append(org_name)
            else:
                active_orgs_skipped.append(org_name)

        elif isinstance(org_config, ScratchOrgConfig):
            runtime.keychain.remove_org(org_name)
            expired_orgs_removed.append(org_name)

    if expired_orgs_removed:
        click.echo(
            f"Successfully removed {len(expired_orgs_removed)} expired scratch orgs: {', '.join(expired_orgs_removed)}"
        )
    else:
        click.echo("No expired scratch orgs to delete. ✨")

    if active_orgs_removed:
        click.echo(
            f"Successfully removed {len(active_orgs_removed)} active scratch orgs: {', '.join(active_orgs_removed)}"
        )
    elif include_active:
        click.echo("No active scratch orgs to delete. ✨")

    if org_shapes_skipped:
        click.echo(f"Skipped org shapes: {', '.join(org_shapes_skipped)}")

    if active_orgs_skipped:
        click.echo(f"Skipped active orgs: {', '.join(active_orgs_skipped)}")


@org.command(name="remove", help="Removes an org from the keychain")
@click.argument("org_name")
@click.option(
    "--global-org",
    is_flag=True,
    help="Set this option to force remove a global org.  Default behavior is to error if you attempt to delete a global org.",
)
@pass_runtime(require_project=False, require_keychain=True)
def org_remove(runtime, org_name, global_org):
    try:
        org_config = runtime.keychain.get_org(org_name)
    except OrgNotFound:
        raise click.ClickException(f"Org {org_name} does not exist in the keychain")

    if org_config.can_delete():
        click.echo("A scratch org was already created, attempting to delete...")
        try:
            org_config.delete_org()
        except Exception as e:
            click.echo("Deleting scratch org failed with error:")
            click.echo(e)

    global_org = global_org or runtime.project_config is None
    runtime.keychain.remove_org(org_name, global_org)


@org.command(
    name="scratch", help="Connects a Salesforce DX Scratch Org to the keychain"
)
@click.argument("config_name")
@click.argument("org_name")
@click.option(
    "--default",
    is_flag=True,
    help="If set, sets the connected org as the new default org",
)
@click.option(
    "--devhub", help="If provided, overrides the devhub used to create the scratch org"
)
@click.option(
    "--days",
    help="If provided, overrides the scratch config default days value for how many days the scratch org should persist",
)
@click.option(
    "--no-password", is_flag=True, help="If set, don't set a password for the org"
)
@pass_runtime(require_keychain=True)
def org_scratch(runtime, config_name, org_name, default, devhub, days, no_password):
    runtime.check_org_overwrite(org_name)

    scratch_configs = getattr(runtime.project_config, "orgs__scratch")
    if not scratch_configs:
        raise click.UsageError("No scratch org configs found in cumulusci.yml")
    scratch_config = scratch_configs.get(config_name)
    if not scratch_config:
        raise click.UsageError(
            f"No scratch org config named {config_name} found in the cumulusci.yml file"
        )

    if devhub:
        scratch_config["devhub"] = devhub

    runtime.keychain.create_scratch_org(
        org_name, config_name, days, set_password=not (no_password)
    )

    if default:
        runtime.keychain.set_default_org(org_name)
        click.echo(f"{org_name} is now the default org")
    else:
        click.echo(f"{org_name} is configured for use")


@org.command(
    name="scratch_delete",
    help="Deletes a Salesforce DX Scratch Org leaving the config in the keychain for regeneration",
)
@click.argument("org_name")
@pass_runtime(require_keychain=True)
def org_scratch_delete(runtime, org_name):
    org_config = runtime.keychain.get_org(org_name)
    if not org_config.scratch:
        raise click.UsageError(f"Org {org_name} is not a scratch org")

    org_config.delete_org()

    org_config.save()


org_shell_cci_help_message = """
The cumulusci shell gives you access to the following objects and functions:

* sf - simple_salesforce connected to your org. [1]
* org_config - local information about your org. [2]
* project_config - information about your project. [3]
* tooling - simple_salesforce connected to the tooling API on your org.
* query() - SOQL query. `help(query)` for more information
* describe() - Inspect object fields. `help(describe)` for more information
* help() - for interactive help on Python
* help(obj) - for help on any specific Python object or module

[1] https://github.com/simple-salesforce/simple-salesforce
[2] https://cumulusci.readthedocs.io/en/latest/api/cumulusci.core.config.html#module-cumulusci.core.config.OrgConfig
[3] https://cumulusci.readthedocs.io/en/latest/api/cumulusci.core.config.html#module-cumulusci.core.config.project_config
"""


class CCIHelp(type(help)):
    def __repr__(self):
        return org_shell_cci_help_message


@org.command(
    name="shell",
    help="Drop into a Python shell with a simple_salesforce connection in `sf`, "
    "as well as the `org_config` and `project_config`.",
)
@click.argument("org_name", required=False)
@click.option("--script", help="Path to a script to run", type=click.Path())
@click.option("--python", help="Python code to run directly")
@pass_runtime(require_keychain=True)
def org_shell(runtime, org_name, script=None, python=None):
    org_name, org_config = runtime.get_org(org_name)
    org_config.refresh_oauth_token(runtime.keychain)

    sf = get_simple_salesforce_connection(runtime.project_config, org_config)
    tooling = get_simple_salesforce_connection(
        runtime.project_config, org_config, base_url="tooling"
    )

    sf_helpers = SimpleSalesforceUIHelpers(sf)

    globals = {
        "sf": sf,
        "tooling": tooling,
        "org_config": org_config,
        "project_config": runtime.project_config,
        "help": CCIHelp(),
        "query": sf_helpers.query,
        "describe": sf_helpers.describe,
    }

    if script:
        if python:
            raise click.UsageError("Cannot specify both --script and --python")
        runpy.run_path(script, init_globals=globals)
    elif python:
        exec(python, globals)
    else:
        code.interact(
            banner=f"Use `sf` to access org `{org_name}` via simple_salesforce\n"
            + "Type `help` for more information about the cci shell.",
            local=globals,
        )

    # Save the org config in case it was modified
    org_config.save()


# Commands for group: task


@task.command(name="list", help="List available tasks for the current context")
@click.option("--plain", is_flag=True, help="Print the table using plain ascii.")
@click.option("--json", "print_json", is_flag=True, help="Print a json string")
@pass_runtime(require_project=False)
def task_list(runtime, plain, print_json):
    tasks = (
        runtime.project_config.list_tasks()
        if runtime.project_config is not None
        else runtime.universal_config.list_tasks()
    )
    plain = plain or runtime.universal_config.cli__plain_output

    if print_json:
        click.echo(json.dumps(tasks))
        return None

    task_groups = group_items(tasks)
    for group, tasks in task_groups.items():
        data = [["Task", "Description"]]
        data.extend(sorted(tasks))
        table = CliTable(data, group, wrap_cols=["Description"])
        table.echo(plain)

    click.echo(
        "Use "
        + click.style("cci task info <task_name>", bold=True)
        + " to get more information about a task."
    )


@task.command(name="doc", help="Exports RST format documentation for all tasks")
@pass_runtime(require_project=False)
def task_doc(runtime):
    config_src = runtime.universal_config

    click.echo("==========================================")
    click.echo("Tasks Reference")
    click.echo("==========================================")
    click.echo("")

    for name, options in config_src.tasks.items():
        task_config = TaskConfig(options)
        doc = doc_task(name, task_config)
        click.echo(doc)
        click.echo("")


@task.command(name="info", help="Displays information for a task")
@click.argument("task_name")
@pass_runtime(require_project=False, require_keychain=True)
def task_info(runtime, task_name):
    task_config = (
        runtime.project_config.get_task(task_name)
        if runtime.project_config is not None
        else runtime.universal_config.get_task(task_name)
    )

    doc = doc_task(task_name, task_config).encode()
    click.echo(rst2ansi(doc))


@task.command(name="run", help="Runs a task")
@click.argument("task_name")
@click.option(
    "--org",
    help="Specify the target org.  By default, runs against the current default org",
)
@click.option(
    "-o",
    nargs=2,
    multiple=True,
    help="Pass task specific options for the task as '-o option value'.  You can specify more than one option by using -o more than once.",
)
@click.option(
    "--debug", is_flag=True, help="Drops into pdb, the Python debugger, on an exception"
)
@click.option(
    "--debug-before",
    is_flag=True,
    help="Drops into the Python debugger right before task start.",
)
@click.option(
    "--debug-after",
    is_flag=True,
    help="Drops into the Python debugger at task completion.",
)
@click.option(
    "--no-prompt",
    is_flag=True,
    help="Disables all prompts.  Set for non-interactive mode use such as calling from scripts or CI systems",
)
@pass_runtime(require_keychain=True)
def task_run(runtime, task_name, org, o, debug, debug_before, debug_after, no_prompt):

    # Get necessary configs
    org, org_config = runtime.get_org(org, fail_if_missing=False)
    task_config = runtime.project_config.get_task(task_name)

    # Get the class to look up options
    class_path = task_config.class_path
    task_class = import_global(class_path)

    # Parse command line options and add to task config
    if o:
        if "options" not in task_config.config:
            task_config.config["options"] = {}
        for name, value in o:
            # Validate the option
            if name not in task_class.task_options:
                raise click.UsageError(
                    f'Option "{name}" is not available for task {task_name}'
                )

            # Override the option in the task config
            task_config.config["options"][name] = value

    # Create and run the task
    try:
        task = task_class(
            task_config.project_config, task_config, org_config=org_config
        )

        if debug_before:
            import pdb

            pdb.set_trace()

        task()

        if debug_after:
            import pdb

            pdb.set_trace()
    finally:
        runtime.alert(f"Task complete: {task_name}")


# Commands for group: flow


@flow.command(name="list", help="List available flows for the current context")
@click.option("--plain", is_flag=True, help="Print the table using plain ascii.")
@click.option("--json", "print_json", is_flag=True, help="Print a json string")
@pass_runtime(require_project=False)
def flow_list(runtime, plain, print_json):
    plain = plain or runtime.universal_config.cli__plain_output
    flows = (
        runtime.project_config.list_flows()
        if runtime.project_config is not None
        else runtime.universal_config.list_flows()
    )

    if print_json:
        click.echo(json.dumps(flows))
        return None

    flow_groups = group_items(flows)
    for group, flows in flow_groups.items():
        data = [["Flow", "Description"]]
        data.extend(sorted(flows))
        table = CliTable(data, group, wrap_cols=["Description"])
        table.echo(plain)

    click.echo(
        "Use "
        + click.style("cci flow info <flow_name>", bold=True)
        + " to get more information about a flow."
    )


@flow.command(name="info", help="Displays information for a flow")
@click.argument("flow_name")
@pass_runtime(require_keychain=True)
def flow_info(runtime, flow_name):
    try:
        coordinator = runtime.get_flow(flow_name)
        output = coordinator.get_summary()
        click.echo(output)
    except FlowNotFoundError as e:
        raise click.UsageError(str(e))


@flow.command(name="run", help="Runs a flow")
@click.argument("flow_name")
@click.option(
    "--org",
    help="Specify the target org.  By default, runs against the current default org",
)
@click.option(
    "--delete-org",
    is_flag=True,
    help="If set, deletes the scratch org after the flow completes",
)
@click.option(
    "--debug", is_flag=True, help="Drops into pdb, the Python debugger, on an exception"
)
@click.option(
    "-o",
    nargs=2,
    multiple=True,
    help="Pass task specific options for the task as '-o taskname__option value'.  You can specify more than one option by using -o more than once.",
)
@click.option(
    "--skip",
    multiple=True,
    help="Specify task names that should be skipped in the flow.  Specify multiple by repeating the --skip option",
)
@click.option(
    "--no-prompt",
    is_flag=True,
    help="Disables all prompts.  Set for non-interactive mode use such as calling from scripts or CI systems",
)
@pass_runtime(require_keychain=True)
def flow_run(runtime, flow_name, org, delete_org, debug, o, skip, no_prompt):

    # Get necessary configs
    org, org_config = runtime.get_org(org)
    if delete_org and not org_config.scratch:
        raise click.UsageError("--delete-org can only be used with a scratch org")

    # Parse command line options
    options = defaultdict(dict)
    if o:
        for key, value in o:
            if "__" in key:
                task_name, option_name = key.split("__")
                options[task_name][option_name] = value
            else:
                raise click.UsageError(
                    "-o option for flows should contain __ to split task name from option name."
                )

    # Create the flow and handle initialization exceptions
    try:
        coordinator = runtime.get_flow(flow_name, options=options)
        coordinator.run(org_config)
    finally:
        runtime.alert(f"Flow Complete: {flow_name}")

    # Delete the scratch org if --delete-org was set
    if delete_org:
        try:
            org_config.delete_org()
        except Exception as e:
            click.echo(
                "Scratch org deletion failed.  Ignoring the error below to complete the flow:"
            )
            click.echo(str(e))


CCI_LOGFILE_PATH = Path.home() / ".cumulusci" / "logs" / "cci.log"


@error.command(
    name="info",
    help="Outputs the most recent traceback (if one exists in the most recent log)",
)
@click.option("--max-lines", "-m", type=int)
def error_info(max_lines: int = 0):
    if not CCI_LOGFILE_PATH.is_file():
        click.echo(f"No logfile found at: {CCI_LOGFILE_PATH}")
    else:
        output = lines_from_traceback(
            CCI_LOGFILE_PATH.read_text(encoding="utf-8"), max_lines
        )
        click.echo(output)


def lines_from_traceback(log_content: str, max_lines: int = 0) -> str:
    """Returns the the last max_lines of the logfile,
    or the whole traceback, whichever is shorter. If
    no stacktrace is found in the logfile, the user is
    notified.
    """
    stacktrace_start = "Traceback (most recent call last):"
    if stacktrace_start not in log_content:
        return f"\nNo stacktrace found in: {CCI_LOGFILE_PATH}\n"

    stacktrace = ""
    for i, line in enumerate(reversed(log_content.split("\n")), 1):
        stacktrace = "\n" + line + stacktrace
        if stacktrace_start in line:
            break
        if i == max_lines:
            break

    return stacktrace


@error.command(name="gist", help="Creates a GitHub gist from the latest logfile")
@pass_runtime(require_project=False, require_keychain=True)
def gist(runtime):
    if CCI_LOGFILE_PATH.is_file():
        log_content = CCI_LOGFILE_PATH.read_text(encoding="utf-8")
    else:
        log_not_found_msg = """No logfile to open at path: {}
        Please ensure you're running this command from the same directory you were experiencing an issue."""
        error_msg = log_not_found_msg.format(CCI_LOGFILE_PATH)
        click.echo(error_msg)
        raise CumulusCIException(error_msg)

    last_cmd_header = "\n\n\nLast Command Run\n================================\n"
    filename = f"cci_output_{datetime.utcnow()}.txt"
    files = {
        filename: {"content": f"{get_context_info()}{last_cmd_header}{log_content}"}
    }

    try:
        gh = RUNTIME.keychain.get_service("github")
        gist = create_gist(
            get_github_api(gh.config["username"], gh.config["password"]),
            "CumulusCI Error Output",
            files,
        )
    except github3.exceptions.NotFoundError:
        raise CumulusCIException(GIST_404_ERR_MSG)
    except Exception as e:
        raise CumulusCIException(
            f"An error occurred attempting to create your gist:\n{e}"
        )
    else:
        click.echo(f"Gist created: {gist.html_url}")
        webbrowser.open(gist.html_url)


def group_items(items):
    groups = defaultdict(list)
    for item in items:
        group_name = item["group"] or "Other"
        groups[group_name].append([item["name"], item["description"]])

    return groups



from collections import defaultdict
from collections import namedtuple
from distutils.version import StrictVersion
import os
import re
from contextlib import contextmanager
from urllib.parse import urlparse
from cumulusci.utils.fileutils import open_fs_resource

import requests
from simple_salesforce import Salesforce

from cumulusci.core.config import BaseConfig
from cumulusci.core.exceptions import CumulusCIException
from cumulusci.core.exceptions import DependencyResolutionError
from cumulusci.core.exceptions import SalesforceCredentialsException
from cumulusci.oauth.salesforce import SalesforceOAuth2
from cumulusci.oauth.salesforce import jwt_session


SKIP_REFRESH = os.environ.get("CUMULUSCI_DISABLE_REFRESH")
SANDBOX_MYDOMAIN_RE = re.compile(r"\.cs\d+\.my\.(.*)salesforce\.com")
MYDOMAIN_RE = re.compile(r"\.my\.(.*)salesforce\.com")


VersionInfo = namedtuple("VersionInfo", ["id", "number"])


class OrgConfig(BaseConfig):
    """ Salesforce org configuration (i.e. org credentials) """

    # make sure it can be mocked for tests
    SalesforceOAuth2 = SalesforceOAuth2

    def __init__(self, config: dict, name: str, keychain=None, global_org=False):
        self.keychain = keychain
        self.global_org = global_org

        self.name = name
        self._community_info_cache = {}
        self._client = None
        self._latest_api_version = None
        self._installed_packages = None
        self._is_person_accounts_enabled = None
        super(OrgConfig, self).__init__(config)

    def refresh_oauth_token(self, keychain, connected_app=None):
        """Get a fresh access token and store it in the org config.

        If the SFDX_CLIENT_ID and SFDX_HUB_KEY environment variables are set,
        this is done using the Oauth2 JWT flow.

        Otherwise it is done using the Oauth2 Refresh Token flow using the connected app
        configured in the keychain's connected_app service.

        Also refreshes user and org info that is cached in the org config.
        """
        # invalidate memoized simple-salesforce client with old token
        self._client = None

        if not SKIP_REFRESH:
            SFDX_CLIENT_ID = os.environ.get("SFDX_CLIENT_ID")
            SFDX_HUB_KEY = os.environ.get("SFDX_HUB_KEY")
            if SFDX_CLIENT_ID and SFDX_HUB_KEY:
                info = jwt_session(
                    SFDX_CLIENT_ID,
                    SFDX_HUB_KEY,
                    self.username,
                    self.instance_url,
                    auth_url=self.id,
                )
            else:
                info = self._refresh_token(keychain, connected_app)
            if info != self.config:
                self.config.update(info)
        self._load_userinfo()
        self._load_orginfo()

    @contextmanager
    def save_if_changed(self):
        orig_config = self.config.copy()
        yield
        if self.config != orig_config:
            self.logger.info("Org info updated, writing to keychain")
            self.save()

    def _refresh_token(self, keychain, connected_app):
        if keychain:  # it might be none'd and caller adds connected_app
            connected_app = keychain.get_service("connected_app")
        if connected_app is None:
            raise AttributeError(
                "No connected app or keychain was passed to refresh_oauth_token."
            )
        client_id = self.client_id
        client_secret = self.client_secret
        if not client_id:
            client_id = connected_app.client_id
            client_secret = connected_app.client_secret
        sf_oauth = self.SalesforceOAuth2(
            client_id,
            client_secret,
            connected_app.callback_url,  # Callback url isn't really used for this call
            auth_site=self.instance_url,
        )

        resp = sf_oauth.refresh_token(self.refresh_token)
        if resp.status_code != 200:
            raise SalesforceCredentialsException(
                f"Error refreshing OAuth token: {resp.text}"
            )
        return resp.json()

    @property
    def lightning_base_url(self):
        instance_url = self.instance_url.rstrip("/")
        if SANDBOX_MYDOMAIN_RE.search(instance_url):
            return SANDBOX_MYDOMAIN_RE.sub(r".lightning.\1force.com", instance_url)
        elif MYDOMAIN_RE.search(instance_url):
            return MYDOMAIN_RE.sub(r".lightning.\1force.com", instance_url)
        else:
            return self.instance_url.split(".")[0] + ".lightning.force.com"

    @property
    def salesforce_client(self):
        if not self._client:
            self._client = Salesforce(
                instance=self.instance_url.replace("https://", ""),
                session_id=self.access_token,
                version=self.latest_api_version,
            )

        return self._client

    @property
    def latest_api_version(self):
        if not self._latest_api_version:
            headers = {"Authorization": "Bearer " + self.access_token}
            response = requests.get(
                self.instance_url + "/services/data", headers=headers
            )
            self._latest_api_version = str(response.json()[-1]["version"])

        return self._latest_api_version

    @property
    def start_url(self):
        start_url = "%s/secur/frontdoor.jsp?sid=%s" % (
            self.instance_url,
            self.access_token,
        )
        return start_url

    @property
    def user_id(self):
        return self.id.split("/")[-1]

    @property
    def org_id(self):
        return self.id.split("/")[-2]

    @property
    def username(self):
        """ Username for the org connection. """
        username = self.config.get("username")
        if not username:
            username = self.userinfo__preferred_username
        return username

    def load_userinfo(self):
        self._load_userinfo()

    def _load_userinfo(self):
        headers = {"Authorization": "Bearer " + self.access_token}
        response = requests.get(
            self.instance_url + "/services/oauth2/userinfo", headers=headers
        )
        if response != self.config.get("userinfo", {}):
            self.config.update({"userinfo": response.json()})

    def can_delete(self):
        return False

    def _load_orginfo(self):
        self._org_sobject = self.salesforce_client.Organization.get(self.org_id)

        result = {
            "org_type": self._org_sobject["OrganizationType"],
            "is_sandbox": self._org_sobject["IsSandbox"],
            "instance_name": self._org_sobject["InstanceName"],
        }
        self.config.update(result)

    @property
    def organization_sobject(self):
        return self._org_sobject

    def _fetch_community_info(self):
        """Use the API to re-fetch information about communities"""
        response = self.salesforce_client.restful("connect/communities")

        # Since community names must be unique, we'll return a dictionary
        # with the community names as keys
        result = {community["name"]: community for community in response["communities"]}
        return result

    def get_community_info(self, community_name, force_refresh=False):
        """Return the community information for the given community

        An API call will be made the first time this function is used,
        and the return values will be cached. Subsequent calls will
        not call the API unless the requested community name is not in
        the cached results, or unless the force_refresh parameter is
        set to True.

        """

        if force_refresh or community_name not in self._community_info_cache:
            self._community_info_cache = self._fetch_community_info()

        if community_name not in self._community_info_cache:
            raise Exception(
                f"Unable to find community information for '{community_name}'"
            )

        return self._community_info_cache[community_name]

    def has_minimum_package_version(self, package_identifier, version_identifier):
        """Return True if the org has a version of the specified package that is
        equal to or newer than the supplied version identifier.

        The package identifier may be either a namespace or a 033 package Id.
        The version identifier should be in "1.2.3" or "1.2.3b4" format.

        A CumulusCIException will be thrown if you request to check a namespace
        and multiple second-generation packages sharing that namespace are installed.
        Use a package Id to handle this circumstance."""
        installed_version = self.installed_packages.get(package_identifier)

        if not installed_version:
            return False
        elif len(installed_version) > 1:
            raise CumulusCIException(
                f"Cannot check installed version of {package_identifier}, because multiple "
                f"packages are installed that match this identifier."
            )

        return installed_version[0].number >= version_identifier

    @property
    def installed_packages(self):
        """installed_packages is a dict mapping a namespace or package Id (033*) to the installed package
        version(s) matching that identifier. All values are lists, because multiple second-generation
        packages may be installed with the same namespace.

        To check if a required package is present, call `has_minimum_package_version()` with either the
        namespace or 033 Id of the desired package and its version, in 1.2.3 format.

        Beta version of a package are represented as "1.2.3b5", where 5 is the build number."""
        if self._installed_packages is None:
            response = self.salesforce_client.restful(
                "tooling/query/?q=SELECT SubscriberPackage.Id, SubscriberPackage.NamespacePrefix, "
                "SubscriberPackageVersion.Id, SubscriberPackageVersion.MajorVersion, "
                "SubscriberPackageVersion.MinorVersion, SubscriberPackageVersion.PatchVersion,  "
                "SubscriberPackageVersion.BuildNumber, SubscriberPackageVersion.IsBeta "
                "FROM InstalledSubscriberPackage"
            )

            self._installed_packages = defaultdict(list)
            for package in response["records"]:
                sp = package["SubscriberPackage"]
                spv = package["SubscriberPackageVersion"]
                version = f"{spv['MajorVersion']}.{spv['MinorVersion']}"
                if spv["PatchVersion"]:
                    version += f".{spv['PatchVersion']}"
                if spv["IsBeta"]:
                    version += f"b{spv['BuildNumber']}"
                version_info = VersionInfo(spv["Id"], StrictVersion(version))
                namespace = sp["NamespacePrefix"]
                self._installed_packages[namespace].append(version_info)
                namespace_version = f"{namespace}@{version}"
                self._installed_packages[namespace_version].append(version_info)
                self._installed_packages[sp["Id"]].append(version_info)

        return self._installed_packages

    def reset_installed_packages(self):
        self._installed_packages = None

    def save(self):
        assert self.keychain, "Keychain was not set on OrgConfig"
        self.keychain.set_org(self, self.global_org)

    def get_domain(self):
        instance_url = self.config.get("instance_url", "")
        return urlparse(instance_url).hostname or ""

    def get_orginfo_cache_dir(self, cachename):
        "Returns a context managed FSResource object"
        assert self.keychain, "Keychain should be set"
        if self.global_org:
            cache_dir = self.keychain.global_config_dir
        else:
            cache_dir = self.keychain.cache_dir
        uniqifier = self.get_domain() + "__" + str(self.username).replace("@", "__")
        cache_dir = cache_dir / "orginfo" / uniqifier / cachename

        cache_dir.mkdir(parents=True, exist_ok=True)
        return open_fs_resource(cache_dir)

    @property
    def is_person_accounts_enabled(self):
        """
        Returns if the org has person accounts enabled, i.e. if Account has an ``IsPersonAccount`` field.

        **Example**

        Selectively run a task in a flow only if Person Accounts is or is not enabled.

        .. code-block:: yaml

            flows:
                load_storytelling_data:
                    steps:
                        1:
                            task: load_dataset
                            options:
                                mapping: datasets/with_person_accounts/mapping.yml
                                sql_path: datasets/with_person_accounts/data.sql
                            when: org_config.is_person_accounts_enabled
                        2:
                            task: load_dataset
                            options:
                                mapping: datasets/without_person_accounts/mapping.yml
                                sql_path: datasets/without_person_accounts/data.sql
                            when: not org_config.is_person_accounts_enabled

        """
        if self._is_person_accounts_enabled is None:
            self._is_person_accounts_enabled = any(
                field["name"] == "IsPersonAccount"
                for field in self.salesforce_client.Account.describe()["fields"]
            )
        return self._is_person_accounts_enabled

    def resolve_04t_dependencies(self, dependencies):
        """Look up 04t SubscriberPackageVersion ids for 1gp project dependencies"""
        new_dependencies = []
        for dependency in dependencies:
            dependency = {**dependency}

            if "namespace" in dependency:
                # get the SubscriberPackageVersion id
                key = f"{dependency['namespace']}@{dependency['version']}"
                version_info = self.installed_packages.get(key)
                if version_info:
                    dependency["version_id"] = version_info[0].id
                else:
                    raise DependencyResolutionError(
                        f"Could not find 04t id for package {key} in org {self.name}"
                    )

            # recurse
            if "dependencies" in dependency:
                dependency["dependencies"] = self.resolve_04t_dependencies(
                    dependency["dependencies"]
                )

            new_dependencies.append(dependency)
        return new_dependencies



"""
This is a library used by locators.robot for testing
custom locator strategies
"""
from cumulusci.robotframework.locator_manager import (
    register_locators,
    translate_locator,
)

locators = {
    # eg: A:breadcrumb:Home
    "breadcrumb": "//span[contains(@class, 'breadcrumbDetail') and text()='{}']"
}


class TestLibraryA:
    ROBOT_LIBRARY_SCOPE = "global"

    def __init__(self):
        register_locators("A", locators)

    def translate_locator(self, prefix, locator):
        return translate_locator(prefix, locator)



import hashlib
import io
import zipfile


def zip_subfolder(zip_src, path):
    if not path.endswith("/"):
        path = path + "/"

    zip_dest = zipfile.ZipFile(io.BytesIO(), "w", zipfile.ZIP_DEFLATED)
    for name in zip_src.namelist():
        if not name.startswith(path):
            continue

        content = zip_src.read(name)
        rel_name = name.replace(path, "", 1)

        if rel_name:
            zip_dest.writestr(rel_name, content)

    return zip_dest


def process_text_in_zipfile(zf, process_file):
    """Process each file in a zip file using the `process_file` function.

    Returns a new zip file.

    `process_file` should be a function which accepts a filename and content as text
    and returns a (possibly modified) filename and content.  The file will be
    replaced with the new content, and renamed if necessary.

    Files with content that cannot be decoded as UTF-8 will be skipped.
    """

    new_zf = zipfile.ZipFile(io.BytesIO(), "w", zipfile.ZIP_DEFLATED)
    for name in zf.namelist():
        content = zf.read(name)
        try:
            content = content.decode("utf-8")
        except UnicodeDecodeError:
            # Probably a binary file; don't change it
            pass
        else:
            name, content = process_file(name, content)
        # writestr handles either bytes or text, and will implicitly encode text as utf-8
        new_zf.writestr(name, content)
    zf.close()
    return new_zf


def hash_zipfile_contents(zf):
    """Returns a hash of a zipfile's file contents.

    Ignores things like file mode, modtime
    """
    h = hashlib.blake2b()
    for name in zf.namelist():
        h.update(name.encode("utf-8"))
        h.update(zf.read(name))
    return h.hexdigest()



import pytest
import os.path

from cumulusci.cli.runtime import CliRuntime
from cumulusci.salesforce_api.utils import get_simple_salesforce_connection
from cumulusci.core.config import TaskConfig


def pytest_addoption(parser, pluginmanager):
    parser.addoption("--org", action="store", default=None, help="org to use")
    parser.addoption(
        "--accelerate-integration-tests",
        action="store_true",
        default=False,
        help="DO run integration tests. Do NOT make calls to a real org. This will error out if you have not run with '--org blah' so that you have cached org output.",
    )


def sf_pytest_orgname(request):
    return request.config.getoption("--org")


@pytest.fixture(scope="session")
def runtime():
    """Get the CumulusCI runtime for the current working directory."""
    return CliRuntime()


@pytest.fixture(scope="session")
def project_config(runtime):
    """Get the project config for the current working directory."""
    return runtime.project_config


@pytest.fixture(scope="session")
def org_config(request, runtime, fallback_orgconfig):
    """Get an org config with an active access token.

    Specify the org name using the --org option when running pytest.
    Or else it will use a dummy org.
    """
    org_name = sf_pytest_orgname(request)
    if org_name:
        org_name, org_config = runtime.get_org(org_name)
        assert org_config.scratch, "You should only run tests against scratch orgs."
        org_config.refresh_oauth_token(runtime.keychain)
    else:
        # fallback_orgconfig can be defined in "conftest" based
        # on the needs of the test suite. For example, for
        # fast running test suites it might return a hardcoded
        # org and for integration test suites it might return
        # a specific default org or throw an exception.
        return fallback_orgconfig()

    return org_config


@pytest.fixture
def sf(request, project_config, org_config):
    """Get a simple-salesforce client for org_config."""
    sf = get_simple_salesforce_connection(project_config, org_config)
    return sf


@pytest.fixture
def create_task(request, project_config, org_config):
    """Get a task _factory_ which can be used to construct task instances."""
    session_project_config = project_config
    session_org_config = org_config

    def create_task(task_class, options=None, project_config=None, org_config=None):
        project_config = project_config or session_project_config
        org_config = org_config or session_org_config
        options = options or {}

        task_config = TaskConfig({"options": options})

        return task_class(project_config, task_config, org_config)

    return create_task


def pytest_configure(config):
    # register an additional marker
    config.addinivalue_line(
        "markers",
        "integration_test(): an integration test that should only be executed when requested",
    )


def pytest_runtest_setup(item):
    is_integration_test = any(item.iter_markers(name="integration_test"))
    if is_integration_test:
        if not item.config.getoption(
            "--accelerate-integration-tests"
        ) and not item.config.getoption("--org"):
            pytest.skip("test requires --org or --accelerate-integration-tests")


@pytest.fixture(scope="module")
def vcr_cassette_dir(request):
    is_integration_test = any(request.node.iter_markers(name="integration_test"))
    test_dir = request.node.fspath.dirname
    if is_integration_test:
        return os.path.join(test_dir, "large_cassettes")  # 8-tracks
    else:  # standard behaviour from
        # https://github.com/ktosiek/pytest-vcr/blob/master/pytest_vcr.py
        return os.path.join(test_dir, "cassettes")



""" Tests for the connectedapp tasks """

from unittest import mock
import os
import pytest
import re
import unittest

try:
    from json.decoder import JSONDecodeError
except ImportError:
    JSONDecodeError = ValueError
from unittest.mock import MagicMock

from cumulusci.core.config import (
    UniversalConfig,
    BaseProjectConfig,
    TaskConfig,
    ServiceConfig,
)
from cumulusci.core.exceptions import TaskOptionsError
from cumulusci.core.keychain import BaseProjectKeychain
from cumulusci.core.keychain import DEFAULT_CONNECTED_APP
from cumulusci.core.tests.utils import MockLoggerMixin
from cumulusci.tasks.connectedapp import CreateConnectedApp
from cumulusci.utils import temporary_dir


class TestCreateConnectedApp(MockLoggerMixin, unittest.TestCase):
    """ Tests for the CreateConnectedApp task """

    def setUp(self):
        self.universal_config = UniversalConfig()
        self.project_config = BaseProjectConfig(
            self.universal_config, config={"noyaml": True}
        )

        keychain = BaseProjectKeychain(self.project_config, "")
        self.project_config.set_keychain(keychain)

        self._task_log_handler.reset()
        self.task_log = self._task_log_handler.messages
        self.base_command = "sfdx force:mdapi:deploy --wait {}".format(
            CreateConnectedApp.deploy_wait
        )
        self.label = "Test_Label"
        self.username = "TestUser@Name"
        self.email = "TestUser@Email"
        self.task_config = TaskConfig(
            {
                "options": {
                    "label": self.label,
                    "username": self.username,
                    "email": self.email,
                }
            }
        )

    def test_init_options(self):
        """ Passed options are correctly initialized """
        self.task_config.config["options"]["connect"] = True
        self.task_config.config["options"]["overwrite"] = True
        task = CreateConnectedApp(self.project_config, self.task_config)
        self.assertEqual(task.options["command"], self.task_config.options__command)
        self.assertEqual(task.options["label"], self.label)
        self.assertEqual(task.options["username"], self.username)
        self.assertEqual(task.options["email"], self.email)
        self.assertIs(task.options["connect"], True)
        self.assertIs(task.options["overwrite"], True)

    def test_init_options_invalid_label(self):
        """ Non-alphanumeric + _ label raises TaskOptionsError """
        self.task_config.config["options"]["label"] = "Test Label"
        with pytest.raises(TaskOptionsError, match="^label value must contain only"):
            CreateConnectedApp(self.project_config, self.task_config)

    def test_init_options_email_default(self):
        """ email option defaults to email from github service """
        del self.task_config.config["options"]["email"]
        self.project_config.config["services"] = {
            "github": {"attributes": {"email": {}}}
        }
        self.project_config.keychain.set_service(
            "github", ServiceConfig({"email": self.email}), True
        )
        task = CreateConnectedApp(self.project_config, self.task_config)
        self.assertEqual(task.options["email"], self.email)

    def test_init_options_email_not_found(self):
        """ TaskOptionsError is raised if no email provided and no github service exists """
        del self.task_config.config["options"]["email"]
        self.project_config.config["services"] = {"github": {"attributes": {}}}
        with pytest.raises(TaskOptionsError, match="github"):
            CreateConnectedApp(self.project_config, self.task_config)

    @mock.patch("cumulusci.tasks.connectedapp.CreateConnectedApp._run_command")
    def test_set_default_username(self, run_command_mock):
        """ _set_default_username calls _run_command """
        task = CreateConnectedApp(self.project_config, self.task_config)
        run_command_mock.side_effect = lambda **kw: kw["output_handler"](
            b'{"result":[{"value":"username"}]}'
        )
        task._set_default_username()
        run_command_mock.assert_called_once()
        self.assertEqual(
            self.task_log["info"], ["Getting username for the default devhub from sfdx"]
        )

    def test_process_json_output(self):
        """ _process_json_output returns valid json """
        task = CreateConnectedApp(self.project_config, self.task_config)
        output = task._process_json_output('{"foo":"bar"}')
        self.assertEqual(output, {"foo": "bar"})

    def test_process_json_output_invalid(self):
        """ _process_json_output with invalid input logs output and raises JSONDecodeError """
        task = CreateConnectedApp(self.project_config, self.task_config)
        with pytest.raises(JSONDecodeError):
            task._process_json_output("invalid")
        self.assertEqual(
            self.task_log["error"], ["Failed to parse json from output: invalid"]
        )

    @mock.patch(
        "cumulusci.tasks.connectedapp.CreateConnectedApp._set_default_username",
        MagicMock(return_value=None),
    )
    def test_process_devhub_output(self):
        """ username is parsed from json response """
        del self.task_config.config["options"]["username"]
        task = CreateConnectedApp(self.project_config, self.task_config)
        task._process_devhub_output('{"result":[{"value":"' + self.username + '"}]}')
        self.assertEqual(task.options.get("username"), self.username)

    @mock.patch(
        "cumulusci.tasks.connectedapp.CreateConnectedApp._set_default_username",
        MagicMock(return_value=None),
    )
    def test_process_devhub_output_not_configured(self):
        """ TaskOptionsError is raised if no username provided and no default found """
        del self.task_config.config["options"]["username"]
        task = CreateConnectedApp(self.project_config, self.task_config)
        with pytest.raises(TaskOptionsError, match="^No sfdx config found"):
            task._process_devhub_output('{"result":[{}]}')

    def test_generate_id_and_secret(self):
        """ client_id and client_secret are generated correctly """
        task = CreateConnectedApp(self.project_config, self.task_config)
        task._generate_id_and_secret()
        self.assertEqual(len(task.client_id), task.client_id_length)
        self.assertEqual(len(task.client_secret), task.client_secret_length)
        self.assertNotEqual(re.match(r"^\w+$", task.client_id), None)
        self.assertNotEqual(re.match(r"^\w+$", task.client_secret), None)

    def test_build_package(self):
        """ tempdir is populated with connected app and package.xml """
        task = CreateConnectedApp(self.project_config, self.task_config)
        with temporary_dir() as tempdir:
            task.tempdir = tempdir
            connected_app_path = os.path.join(
                task.tempdir, "connectedApps", "{}.connectedApp".format(self.label)
            )
            task._build_package()
            self.assertTrue(os.path.isdir(os.path.join(task.tempdir, "connectedApps")))
            self.assertTrue(os.path.isfile(os.path.join(task.tempdir, "package.xml")))
            self.assertTrue(os.path.isfile(connected_app_path))
            with open(connected_app_path, "r") as f:
                connected_app = f.read()
                self.assertTrue("<label>{}<".format(self.label) in connected_app)
                self.assertTrue("<contactEmail>{}<".format(self.email) in connected_app)
                self.assertTrue(
                    "<consumerKey>{}<".format(task.client_id) in connected_app
                )
                self.assertTrue(
                    "<consumerSecret>{}<".format(task.client_secret) in connected_app
                )

    def test_connect_service(self):
        """ connected app gets added to the keychain connected_app service """
        self.project_config.config["services"] = {
            "connected_app": {
                "attributes": {"callback_url": {}, "client_id": {}, "client_secret": {}}
            }
        }
        task = CreateConnectedApp(self.project_config, self.task_config)
        task._connect_service()
        connected_app = self.project_config.keychain.get_service("connected_app")
        self.assertEqual(connected_app.callback_url, "http://localhost:8080/callback")
        self.assertEqual(connected_app.client_id, task.client_id)
        self.assertEqual(connected_app.client_secret, task.client_secret)

    def test_validate_service_overwrite_false(self):
        """ attempting to overwrite connected_app service without overwrite = True fails """
        self.project_config.config["services"] = {
            "connected_app": {
                "attributes": {"callback_url": {}, "client_id": {}, "client_secret": {}}
            }
        }
        self.project_config.keychain.set_service(
            "connected_app",
            ServiceConfig(
                {
                    "callback_url": "http://callback",
                    "client_id": "ClientId",
                    "client_secret": "ClientSecret",
                }
            ),
            True,
        )
        task = CreateConnectedApp(self.project_config, self.task_config)
        with pytest.raises(
            TaskOptionsError, match="^The CumulusCI keychain already contains"
        ):
            task._validate_connect_service()

    @mock.patch("cumulusci.tasks.sfdx.SFDXBaseTask._run_task")
    def test_run_task(self, run_task_mock):
        """ _run_task formats command, calls SFDXBaseTask._run_task, and does not connect service by default """
        self.project_config.config["services"] = {
            "connected_app": {
                "attributes": {"callback_url": {}, "client_id": {}, "client_secret": {}}
            }
        }
        task = CreateConnectedApp(self.project_config, self.task_config)
        task._run_task()
        run_task_mock.assert_called_once()
        self.assertFalse(os.path.isdir(task.tempdir))
        connected_app = self.project_config.keychain.get_service("connected_app")
        assert connected_app is DEFAULT_CONNECTED_APP

    @mock.patch("cumulusci.tasks.sfdx.SFDXBaseTask._run_task")
    @mock.patch("cumulusci.tasks.connectedapp.CreateConnectedApp._connect_service")
    def test_run_task_connect(self, run_task_mock, connect_service_mock):
        """ _run_task calls _connect_service if connect option is True """
        self.project_config.config["services"] = {
            "connected_app": {
                "attributes": {"callback_url": {}, "client_id": {}, "client_secret": {}}
            }
        }
        self.task_config.config["options"]["connect"] = True
        task = CreateConnectedApp(self.project_config, self.task_config)
        task._run_task()
        run_task_mock.assert_called_once()
        connect_service_mock.assert_called_once()



import json
import unittest
from datetime import datetime
from unittest import mock

import responses
from simple_salesforce.exceptions import SalesforceMalformedRequest

from cumulusci.core.exceptions import SalesforceException
from cumulusci.tasks.salesforce import CreateCommunity

from .util import create_task

task_options = {
    "name": "Test Community",
    "description": "Community Details",
    "template": "VF Template",
    "url_path_prefix": "test",
    "skip_existing": False,
}
task_options_no_url_path_prefix = {
    "name": "Test Community",
    "description": "Community Details",
    "template": "VF Template",
    "skip_existing": False,
}


class test_CreateCommunity(unittest.TestCase):
    @responses.activate
    def test_creates_community(self):
        cc_task = create_task(CreateCommunity, task_options)
        community_url = "{}/services/data/v48.0/connect/communities".format(
            cc_task.org_config.instance_url
        )

        responses.add(
            method=responses.GET,
            url=community_url,
            status=200,
            json={"communities": []},
        )
        responses.add(method=responses.POST, url=community_url, status=200, json={})
        responses.add(
            method=responses.GET,
            url=community_url,
            status=200,
            json={"communities": [{"name": "Test Community", "id": "000000000000000"}]},
        )

        cc_task()

        self.assertEqual(3, len(responses.calls))
        self.assertEqual(community_url, responses.calls[1].request.url)
        self.assertEqual(community_url, responses.calls[2].request.url)
        self.assertEqual(
            json.dumps(
                {
                    "name": "Test Community",
                    "description": "Community Details",
                    "templateName": "VF Template",
                    "urlPathPrefix": "test",
                }
            ),
            responses.calls[1].request.body,
        )

    @responses.activate
    def test_creates_community_no_url_path_prefix(self):
        cc_task = create_task(CreateCommunity, task_options_no_url_path_prefix)
        community_url = "{}/services/data/v48.0/connect/communities".format(
            cc_task.org_config.instance_url
        )

        responses.add(
            method=responses.GET,
            url=community_url,
            status=200,
            json={"communities": []},
        )
        responses.add(method=responses.POST, url=community_url, status=200, json={})
        responses.add(
            method=responses.GET,
            url=community_url,
            status=200,
            json={"communities": [{"name": "Test Community", "id": "000000000000000"}]},
        )

        cc_task()

        self.assertEqual(3, len(responses.calls))
        self.assertEqual(community_url, responses.calls[1].request.url)
        self.assertEqual(community_url, responses.calls[2].request.url)
        self.assertEqual(
            json.dumps(
                {
                    "name": "Test Community",
                    "description": "Community Details",
                    "templateName": "VF Template",
                    "urlPathPrefix": "",
                }
            ),
            responses.calls[1].request.body,
        )

    @responses.activate
    def test_error_for_existing_community(self):
        cc_task = create_task(CreateCommunity, task_options)
        community_url = "{}/services/data/v48.0/connect/communities".format(
            cc_task.org_config.instance_url
        )

        responses.add(
            method=responses.GET,
            url=community_url,
            status=200,
            json={"communities": [{"name": "Test Community", "id": "000000000000000"}]},
        )

        with self.assertRaises(Exception):
            cc_task()

    @responses.activate
    def test_no_error_for_existing_community_when_skip_existing(self):
        task_options["skip_existing"] = True
        cc_task = create_task(CreateCommunity, task_options)
        community_url = "{}/services/data/v48.0/connect/communities".format(
            cc_task.org_config.instance_url
        )

        responses.add(
            method=responses.GET,
            url=community_url,
            status=200,
            json={"communities": [{"name": "Test Community", "id": "000000000000000"}]},
        )

        cc_task()

    @responses.activate
    def test_handles_community_created_between_tries(self):
        cc_task = create_task(CreateCommunity, task_options)
        community_url = "{}/services/data/v48.0/connect/communities".format(
            cc_task.org_config.instance_url
        )

        responses.add(
            method=responses.GET,
            url=community_url,
            status=200,
            json={"communities": []},
        )
        responses.add(
            method=responses.POST,
            url=community_url,
            status=400,
            json=[
                {
                    "errorCode": "INVALID_INPUT",
                    "message": "Error: A Community with this name already exists.",
                }
            ],
        )
        responses.add(
            method=responses.GET,
            url=community_url,
            status=200,
            json={"communities": [{"name": "Test Community", "id": "000000000000000"}]},
        )

        cc_task()

    @responses.activate
    def test_throws_exception_for_existing_url_path_prefix(self):
        cc_task = create_task(CreateCommunity, task_options)
        community_url = "{}/services/data/v48.0/connect/communities".format(
            cc_task.org_config.instance_url
        )

        responses.add(
            method=responses.GET,
            url=community_url,
            status=200,
            json={"communities": []},
        )
        responses.add(
            method=responses.POST,
            url=community_url,
            status=400,
            json=[
                {
                    "errorCode": "INVALID_INPUT",
                    "message": "That URL is already in use. Please enter a unique one.",
                }
            ],
        )

        cc_task._init_task()
        with self.assertRaises(SalesforceMalformedRequest):
            cc_task._run_task()()

    @responses.activate
    def test_throws_exception_for_existing_no_url_path_prefix(self):
        cc_task = create_task(CreateCommunity, task_options_no_url_path_prefix)
        community_url = "{}/services/data/v48.0/connect/communities".format(
            cc_task.org_config.instance_url
        )

        responses.add(
            method=responses.GET,
            url=community_url,
            status=200,
            json={"communities": []},
        )
        responses.add(
            method=responses.POST,
            url=community_url,
            status=400,
            json=[
                {
                    "errorCode": "INVALID_INPUT",
                    "message": "Enter a URL for your community.",
                }
            ],
        )

        cc_task._init_task()
        with self.assertRaises(SalesforceMalformedRequest):
            cc_task._run_task()()

    @responses.activate
    def test_waits_for_community_result__not_complete(self):
        cc_task = create_task(CreateCommunity, task_options)

        community_url = "{}/services/data/v48.0/connect/communities".format(
            cc_task.org_config.instance_url
        )
        responses.add(
            method=responses.GET,
            url=community_url,
            status=200,
            json={"communities": []},
        )

        cc_task._init_task()
        cc_task.time_start = datetime.now()
        cc_task._poll_action()

        self.assertFalse(cc_task.poll_complete)

    @responses.activate
    def test_waits_for_community_result__complete(self):
        cc_task = create_task(CreateCommunity, task_options)
        community_url = "{}/services/data/v48.0/connect/communities".format(
            cc_task.org_config.instance_url
        )
        responses.add(
            method=responses.GET,
            url=community_url,
            status=200,
            json={"communities": [{"name": "Test Community", "id": "000000000000000"}]},
        )
        cc_task.logger = mock.Mock()

        cc_task._init_task()
        cc_task.time_start = datetime.now()
        cc_task._poll_action()

        self.assertTrue(cc_task.poll_complete)
        cc_task.logger.info.assert_called_once_with("Community 000000000000000 created")

    def test_throws_exception_for_timeout(self):
        cc_task = create_task(CreateCommunity, task_options)

        cc_task.time_start = datetime(2019, 1, 1)
        with self.assertRaises(SalesforceException):
            cc_task._poll_action()



import datetime
import json

from cumulusci.core.config import OrgConfig
from cumulusci.core.exceptions import SfdxOrgException
from cumulusci.core.sfdx import sfdx
from cumulusci.utils import get_git_config

nl = "\n"  # fstrings can't contain backslashes


class SfdxOrgConfig(OrgConfig):
    """Org config which loads from sfdx keychain"""

    @property
    def sfdx_info(self):
        if hasattr(self, "_sfdx_info"):
            return self._sfdx_info

        # On-demand creation of scratch orgs
        if self.create_org is not None and not self.created:
            self.create_org()

        username = self.config.get("username")
        assert username is not None, "SfdxOrgConfig must have a username"

        self.logger.info(f"Getting org info from Salesforce CLI for {username}")

        # Call force:org:display and parse output to get instance_url and
        # access_token
        p = sfdx("force:org:display --json", self.username)

        org_info = None
        stderr_list = [line.strip() for line in p.stderr_text]
        stdout_list = [line.strip() for line in p.stdout_text]

        if p.returncode:
            self.logger.error(f"Return code: {p.returncode}")
            for line in stderr_list:
                self.logger.error(line)
            for line in stdout_list:
                self.logger.error(line)
            message = f"\nstderr:\n{nl.join(stderr_list)}"
            message += f"\nstdout:\n{nl.join(stdout_list)}"
            raise SfdxOrgException(message)

        else:
            try:
                org_info = json.loads("".join(stdout_list))
            except Exception as e:
                raise SfdxOrgException(
                    "Failed to parse json from output.\n  "
                    f"Exception: {e.__class__.__name__}\n  Output: {''.join(stdout_list)}"
                )
            org_id = org_info["result"]["accessToken"].split("!")[0]

        sfdx_info = {
            "instance_url": org_info["result"]["instanceUrl"],
            "access_token": org_info["result"]["accessToken"],
            "org_id": org_id,
            "username": org_info["result"]["username"],
        }
        if org_info["result"].get("password"):
            sfdx_info["password"] = org_info["result"]["password"]
        self._sfdx_info = sfdx_info
        self._sfdx_info_date = datetime.datetime.utcnow()
        self.config.update(sfdx_info)

        sfdx_info.update(
            {
                "created_date": org_info["result"].get("createdDate"),
                "expiration_date": org_info["result"].get("expirationDate"),
            }
        )
        return sfdx_info

    @property
    def access_token(self):
        return self.sfdx_info["access_token"]

    @property
    def instance_url(self):
        return self.config.get("instance_url") or self.sfdx_info["instance_url"]

    @property
    def org_id(self):
        org_id = self.config.get("org_id")
        if not org_id:
            org_id = self.sfdx_info["org_id"]
        return org_id

    @property
    def user_id(self):
        if not self.config.get("user_id"):
            result = self.salesforce_client.query_all(
                f"SELECT Id FROM User WHERE UserName='{self.username}'"
            )
            self.config["user_id"] = result["records"][0]["Id"]
        return self.config["user_id"]

    @property
    def username(self):
        username = self.config.get("username")
        if not username:
            username = self.sfdx_info["username"]
        return username

    @property
    def password(self):
        password = self.config.get("password")
        if not password:
            password = self.sfdx_info["password"]
        return password

    @property
    def email_address(self):
        email_address = self.config.get("email_address")
        if not email_address:
            email_address = get_git_config("user.email")
            self.config["email_address"] = email_address

        return email_address

    def force_refresh_oauth_token(self):
        # Call force:org:display and parse output to get instance_url and
        # access_token
        p = sfdx("force:org:open -r", self.username, log_note="Refreshing OAuth token")

        stdout_list = [line.strip() for line in p.stdout_text]

        if p.returncode:
            self.logger.error(f"Return code: {p.returncode}")
            for line in stdout_list:
                self.logger.error(line)
            message = f"Message: {nl.join(stdout_list)}"
            raise SfdxOrgException(message)

    def refresh_oauth_token(self, keychain):
        """ Use sfdx force:org:describe to refresh token instead of built in OAuth handling """
        self._client = None
        if hasattr(self, "_sfdx_info"):
            # Cache the sfdx_info for 1 hour to avoid unnecessary calls out to sfdx CLI
            delta = datetime.datetime.utcnow() - self._sfdx_info_date
            if delta.total_seconds() > 3600:
                del self._sfdx_info

                # Force a token refresh
                self.force_refresh_oauth_token()

        # Get org info via sfdx force:org:display
        self.sfdx_info
        # Get additional org info by querying API
        self._load_orginfo()



import http.client
import io
import unittest
from collections import defaultdict
from xml.dom.minidom import parseString
import datetime

from requests import Response
import responses
import pytest

from cumulusci.tests.util import create_project_config
from cumulusci.tests.util import DummyOrgConfig
from cumulusci.core.config import TaskConfig
from cumulusci.core.exceptions import ApexTestException
from cumulusci.core.tasks import BaseTask
from cumulusci.salesforce_api.exceptions import MetadataApiError
from cumulusci.salesforce_api.exceptions import MetadataParseError
from cumulusci.salesforce_api.exceptions import MetadataComponentFailure
from cumulusci.salesforce_api.metadata import BaseMetadataApiCall
from cumulusci.salesforce_api.metadata import ApiDeploy
from cumulusci.salesforce_api.metadata import ApiListMetadata
from cumulusci.salesforce_api.metadata import ApiRetrieveUnpackaged
from cumulusci.salesforce_api.metadata import ApiRetrieveInstalledPackages
from cumulusci.salesforce_api.metadata import ApiRetrievePackaged
from cumulusci.salesforce_api.package_zip import BasePackageZipBuilder
from cumulusci.salesforce_api.package_zip import CreatePackageZipBuilder
from cumulusci.salesforce_api.package_zip import InstallPackageZipBuilder
from cumulusci.salesforce_api.tests.metadata_test_strings import deploy_status_envelope
from cumulusci.salesforce_api.tests.metadata_test_strings import deploy_result
from cumulusci.salesforce_api.tests.metadata_test_strings import deploy_result_failure
from cumulusci.salesforce_api.tests.metadata_test_strings import (
    list_metadata_start_envelope,
)
from cumulusci.salesforce_api.tests.metadata_test_strings import list_metadata_result
from cumulusci.salesforce_api.tests.metadata_test_strings import (
    list_metadata_result_bad_val,
)
from cumulusci.salesforce_api.tests.metadata_test_strings import (
    retrieve_packaged_start_envelope,
)
from cumulusci.salesforce_api.tests.metadata_test_strings import (
    retrieve_unpackaged_start_envelope,
)
from cumulusci.salesforce_api.tests.metadata_test_strings import retrieve_result
from cumulusci.salesforce_api.tests.metadata_test_strings import result_envelope
from cumulusci.salesforce_api.tests.metadata_test_strings import status_envelope


class DummyPackageZipBuilder(BasePackageZipBuilder):
    def _populate_zip(self):
        return


class BaseTestMetadataApi(unittest.TestCase):
    api_class = BaseMetadataApiCall
    envelope_start = None
    envelope_status = status_envelope
    envelope_result = result_envelope

    def setUp(self):

        # Set up the mock values
        self.repo_name = "TestRepo"
        self.repo_owner = "TestOwner"
        self.repo_api_url = "https://api.github.com/repos/{}/{}".format(
            self.repo_owner, self.repo_name
        )
        self.branch = "main"

        # Create the project config
        self.project_config = create_project_config(self.repo_name, self.repo_owner)

        if not self.envelope_start:
            self.envelope_start = self.api_class.soap_envelope_start

    def _create_task(self, task_config=None, org_config=None):
        if not task_config:
            task_config = {}
        if not org_config:
            org_config = {}
        task = BaseTask(
            project_config=self.project_config,
            task_config=TaskConfig(task_config),
            org_config=DummyOrgConfig(org_config),
        )
        return task

    def _mock_call_mdapi(self, api, response, status_code=None):
        if not status_code:
            status_code = 200
        responses.add(
            method=responses.POST,
            url=api._build_endpoint_url(),
            body=response,
            status=status_code,
            content_type="text/xml; charset=utf-8",
        )
        return response

    def _create_instance(self, task, api_version=None):
        return self.api_class(task, api_version=api_version)

    def test_init(self):
        task = self._create_task()
        api = self._create_instance(task)
        self.assertEqual(api.task, task)
        self.assertEqual(
            api.api_version, self.project_config.project__package__api_version
        )

    def test_build_endpoint_url(self):
        org_config = {
            "instance_url": "https://na12.salesforce.com",
            "id": "https://login.salesforce.com/id/00D000000000000ABC/005000000000000ABC",
        }
        task = self._create_task(org_config=org_config)
        api = self._create_instance(task)
        self.assertEqual(
            api._build_endpoint_url(),
            "{}/services/Soap/m/{}/{}".format(
                org_config["instance_url"],
                self.project_config.project__package__api_version,
                task.org_config.org_id,
            ),
        )

    def test_build_endpoint_url_mydomain(self):
        org_config = {
            "instance_url": "https://test-org.na12.my.salesforce.com",
            "id": "https://login.salesforce.com/id/00D000000000000ABC/005000000000000ABC",
        }
        task = self._create_task(org_config=org_config)
        api = self._create_instance(task)
        self.assertEqual(
            api._build_endpoint_url(),
            "https://na12.salesforce.com/services/Soap/m/{}/{}".format(
                self.project_config.project__package__api_version,
                task.org_config.org_id,
            ),
        )

    def test_build_endpoint_url_apiversion(self):
        org_config = {
            "instance_url": "https://na12.salesforce.com",
            "id": "https://login.salesforce.com/id/00D000000000000ABC/005000000000000ABC",
        }
        task = self._create_task(org_config=org_config)
        api_version = "43.0"
        api = self._create_instance(task, api_version=api_version)
        self.assertEqual(
            api._build_endpoint_url(),
            "{}/services/Soap/m/{}/{}".format(
                org_config["instance_url"], api_version, task.org_config.org_id
            ),
        )

    def test_build_envelope_result(self):
        task = self._create_task()
        api = self._create_instance(task)
        if not self.api_class.soap_envelope_result:
            api.soap_envelope_result = "{process_id}"
            expected = "123"
        else:
            expected = self.envelope_result.format(process_id="123")
        api.process_id = "123"
        self.assertEqual(api._build_envelope_result(), expected)

    def test_build_envelope_start(self):
        task = self._create_task()
        api = self._create_instance(task)
        if not self.api_class.soap_envelope_start:
            api.soap_envelope_start = "{api_version}"
            expected = str(self.project_config.project__package__api_version)
        else:
            expected = self._expected_envelope_start()

        self.assertEqual(api._build_envelope_start(), expected)

    def _expected_envelope_start(self):
        return self.envelope_start.format(
            api_version=self.project_config.project__package__api_version
        )

    def test_build_envelope_status(self):
        task = self._create_task()
        api = self._create_instance(task)
        process_id = "123"
        if not self.api_class.soap_envelope_status:
            api.soap_envelope_status = "{process_id}"
            expected = process_id
        else:
            expected = self.envelope_status.format(process_id=process_id)
        api.process_id = process_id
        self.assertEqual(api._build_envelope_status(), expected)

    def test_build_headers(self):
        action = "foo"
        message = "12345678"
        task = self._create_task()
        api = self._create_instance(task)
        self.assertEqual(
            api._build_headers(action, message),
            {
                u"Content-Type": u"text/xml; charset=UTF-8",
                u"Content-Length": "8",
                u"SOAPAction": u"foo",
            },
        )

    @responses.activate
    def test_call_faultcode(self):
        org_config = {
            "instance_url": "https://na12.salesforce.com",
            "id": "https://login.salesforce.com/id/00D000000000000ABC/005000000000000ABC",
            "access_token": "0123456789",
        }
        task = self._create_task(org_config=org_config)
        api = self._create_instance(task)
        if not self.api_class.soap_envelope_start:
            api.soap_envelope_start = "{api_version}"
        response = '<?xml version="1.0" encoding="UTF-8"?><faultcode>foo</faultcode>'
        self._mock_call_mdapi(api, response)
        with self.assertRaises(MetadataApiError):
            api()

    @responses.activate
    def test_call_success(self):
        org_config = {
            "instance_url": "https://na12.salesforce.com",
            "id": "https://login.salesforce.com/id/00D000000000000ABC/005000000000000ABC",
            "access_token": "0123456789",
        }
        task = self._create_task(org_config=org_config)
        api = self._create_instance(task)
        if not self.api_class.soap_envelope_start:
            api.soap_envelope_start = "{api_version}"
        if not self.api_class.soap_envelope_status:
            api.soap_envelope_status = "{process_id}"
        if not self.api_class.soap_envelope_result:
            api.soap_envelope_result = "{process_id}"

        response = '<?xml version="1.0" encoding="UTF-8"?><id>1234567890</id>'
        self._mock_call_mdapi(api, response)
        response_status = '<?xml version="1.0" encoding="UTF-8"?><done>true</done>'
        self._mock_call_mdapi(api, response_status)
        response_result = '<?xml version="1.0" encoding="UTF-8"?><foo>bar</foo>'
        response_result = self._response_call_success_result(response_result)
        self._mock_call_mdapi(api, response_result)

        resp = api()
        expected_resp = self._expected_call_success_result(response_result)

        self.assertEqual(resp, expected_resp)

    def _expected_call_success_result(self, response_result):
        return response_result

    def _response_call_success_result(self, response_result):
        return response_result

    def test_get_element_value(self):
        task = self._create_task()
        api = self._create_instance(task)
        dom = parseString("<foo>bar</foo>")
        self.assertEqual(api._get_element_value(dom, "foo"), "bar")

    def test_get_element_value_not_found(self):
        task = self._create_task()
        api = self._create_instance(task)
        dom = parseString("<foo>bar</foo>")
        self.assertEqual(api._get_element_value(dom, "baz"), None)

    def test_get_element_value_empty(self):
        task = self._create_task()
        api = self._create_instance(task)
        dom = parseString("<foo />")
        self.assertEqual(api._get_element_value(dom, "foo"), None)

    def test_get_check_interval(self):
        task = self._create_task()
        api = self._create_instance(task)
        api.check_num = 1
        self.assertEqual(api._get_check_interval(), 1)
        api.check_num = 10
        self.assertEqual(api._get_check_interval(), 4)

    @responses.activate
    def test_get_response_faultcode(self):
        org_config = {
            "instance_url": "https://na12.salesforce.com",
            "id": "https://login.salesforce.com/id/00D000000000000ABC/005000000000000ABC",
            "access_token": "0123456789",
        }
        task = self._create_task(org_config=org_config)
        api = self._create_instance(task)
        if not self.api_class.soap_envelope_start:
            api.soap_envelope_start = "{api_version}"
        response = '<?xml version="1.0" encoding="UTF-8"?><faultcode>foo</faultcode>'
        self._mock_call_mdapi(api, response)
        with self.assertRaises(MetadataApiError):
            api._get_response()

    @responses.activate
    def test_get_response_faultcode_and_string(self):
        org_config = {
            "instance_url": "https://na12.salesforce.com",
            "id": "https://login.salesforce.com/id/00D000000000000ABC/005000000000000ABC",
            "access_token": "0123456789",
        }
        task = self._create_task(org_config=org_config)
        api = self._create_instance(task)
        if not self.api_class.soap_envelope_start:
            api.soap_envelope_start = "{api_version}"
        response = '<?xml version="1.0" encoding="UTF-8"?>'
        response += "\n<test>"
        response += "\n  <faultcode>foo</faultcode>"
        response += "\n  <faultstring>bar</faultstring>"
        response += "\n</test>"
        self._mock_call_mdapi(api, response)
        with self.assertRaises(MetadataApiError):
            api._get_response()

    @responses.activate
    def test_get_response_faultcode_invalid_session_no_refresh(self):
        org_config = {
            "instance_url": "https://na12.salesforce.com",
            "id": "https://login.salesforce.com/id/00D000000000000ABC/005000000000000ABC",
            "access_token": "0123456789",
        }
        task = self._create_task(org_config=org_config)
        api = self._create_instance(task)
        if not self.api_class.soap_envelope_start:
            api.soap_envelope_start = "{api_version}"
        response = '<?xml version="1.0" encoding="UTF-8"?><faultcode>sf:INVALID_SESSION_ID</faultcode>'
        self._mock_call_mdapi(api, response)
        with self.assertRaises(MetadataApiError):
            api._get_response()
        self.assertEqual(api.status, "Failed")

    @responses.activate
    def test_get_response_faultcode_invalid_session_refresh(self):
        org_config = {
            "instance_url": "https://na12.salesforce.com",
            "id": "https://login.salesforce.com/id/00D000000000000ABC/005000000000000ABC",
            "access_token": "0123456789",
            "refresh_token": "abcdefghij",
        }
        task = self._create_task(org_config=org_config)
        api = self._create_instance(task)
        if not self.api_class.soap_envelope_start:
            api.soap_envelope_start = "{api_version}"
        if not self.api_class.soap_envelope_status:
            api.soap_envelope_status = "{process_id}"

        mock_responses = []
        mock_responses.append(b'<?xml version="1.0" encoding="UTF-8"?><id>123</id>')
        mock_responses.append(
            b'<?xml version="1.0" encoding="UTF-8"?><faultcode>sf:INVALID_SESSION_ID</faultcode>'
        )
        mock_responses.append(b'<?xml version="1.0" encoding="UTF-8"?><foo>bar</foo>')
        for response in mock_responses:
            self._mock_call_mdapi(api, response)

        resp = api._get_response()
        self.assertEqual(resp.content, mock_responses[2])

    @responses.activate
    def test_get_response_start_error_500(self):
        org_config = {
            "instance_url": "https://na12.salesforce.com",
            "id": "https://login.salesforce.com/id/00D000000000000ABC/005000000000000ABC",
            "access_token": "0123456789",
        }
        status_code = http.client.INTERNAL_SERVER_ERROR  # HTTP Error 500
        task = self._create_task(org_config=org_config)
        api = self._create_instance(task)
        if not self.api_class.soap_envelope_start:
            api.soap_envelope_start = "{api_version}"
        if not self.api_class.soap_envelope_status:
            api.soap_envelope_status = "{process_id}"

        response = '<?xml version="1.0" encoding="UTF-8"?><foo>start</foo>'
        self._mock_call_mdapi(api, response, status_code)

        with self.assertRaises(MetadataApiError):
            api._get_response()

    @responses.activate
    def test_get_response_status_error_500(self):
        org_config = {
            "instance_url": "https://na12.salesforce.com",
            "id": "https://login.salesforce.com/id/00D000000000000ABC/005000000000000ABC",
            "access_token": "0123456789",
        }
        status_code = http.client.INTERNAL_SERVER_ERROR  # HTTP Error 500
        task = self._create_task(org_config=org_config)
        api = self._create_instance(task)
        if not self.api_class.soap_envelope_start:
            api.soap_envelope_start = "{api_version}"
        if not self.api_class.soap_envelope_status:
            api.soap_envelope_status = "{process_id}"

        response = '<?xml version="1.0" encoding="UTF-8"?><id>1234567890</id>'
        self._mock_call_mdapi(api, response)
        self._mock_call_mdapi(api, response, status_code)

        with self.assertRaises(MetadataApiError):
            api._get_response()

    @responses.activate
    def test_get_response_status_no_loop(self):
        org_config = {
            "instance_url": "https://na12.salesforce.com",
            "id": "https://login.salesforce.com/id/00D000000000000ABC/005000000000000ABC",
            "access_token": "0123456789",
        }
        task = self._create_task(org_config=org_config)
        api = self._create_instance(task)
        if not self.api_class.soap_envelope_start:
            api.soap_envelope_start = "{api_version}"
        if not self.api_class.soap_envelope_status:
            api.soap_envelope_status = "{process_id}"
        if not self.api_class.soap_envelope_result:
            api.soap_envelope_result = "{process_id}"

        response = b'<?xml version="1.0" encoding="UTF-8"?><id>1234567890</id>'
        self._mock_call_mdapi(api, response)
        response_status = b'<?xml version="1.0" encoding="UTF-8"?><done>true</done>'
        self._mock_call_mdapi(api, response_status)
        response_result = b'<?xml version="1.0" encoding="UTF-8"?><foo>bar</foo>'
        self._mock_call_mdapi(api, response_result)

        resp = api._get_response()

        self.assertEqual(resp.content, response_result)

    @responses.activate
    def test_get_response_status_loop_twice(self):
        org_config = {
            "instance_url": "https://na12.salesforce.com",
            "id": "https://login.salesforce.com/id/00D000000000000ABC/005000000000000ABC",
            "access_token": "0123456789",
        }
        task = self._create_task(org_config=org_config)
        api = self._create_instance(task)
        if not self.api_class.soap_envelope_start:
            api.soap_envelope_start = "{api_version}"
        if not self.api_class.soap_envelope_status:
            api.soap_envelope_status = "{process_id}"
        if not self.api_class.soap_envelope_result:
            api.soap_envelope_result = "{process_id}"
        api.check_interval = 0

        response = b'<?xml version="1.0" encoding="UTF-8"?><id>1234567890</id>'
        self._mock_call_mdapi(api, response)
        response_status = b'<?xml version="1.0" encoding="UTF-8"?><done>false</done>'
        self._mock_call_mdapi(api, response_status)
        response_status = b'<?xml version="1.0" encoding="UTF-8"?><done>false</done>'
        self._mock_call_mdapi(api, response_status)
        response_status = b'<?xml version="1.0" encoding="UTF-8"?><done>true</done>'
        self._mock_call_mdapi(api, response_status)
        response_result = b'<?xml version="1.0" encoding="UTF-8"?><foo>bar</foo>'
        self._mock_call_mdapi(api, response_result)

        resp = api._get_response()

        self.assertEqual(resp.content, response_result)

        self.assertEqual(api.status, "Done")

        self.assertEqual(api.check_num, 4)

    def test_process_response_status_no_done_element(self):
        task = self._create_task()
        api = self._create_instance(task)
        response = Response()
        response.status_code = 200
        response.raw = io.BytesIO(
            b'<?xml version="1.0" encoding="UTF-8"?><foo>status</foo>'
        )
        res = api._process_response_status(response)
        self.assertEqual(api.status, "Failed")
        self.assertEqual(res.content, response.content)

    def test_process_response_status_done_is_true(self):
        task = self._create_task()
        api = self._create_instance(task)
        response = Response()
        response.status_code = 200
        response.raw = io.BytesIO(
            b'<?xml version="1.0" encoding="UTF-8"?><done>true</done>'
        )
        res = api._process_response_status(response)
        self.assertEqual(api.status, "Done")
        self.assertEqual(res.content, response.content)

    def test_process_response_status_pending(self):
        task = self._create_task()
        api = self._create_instance(task)
        response = Response()
        response.status_code = 200
        response.raw = io.BytesIO(
            b'<?xml version="1.0" encoding="UTF-8"?><done>false</done>'
        )
        res = api._process_response_status(response)
        self.assertEqual(api.status, "Pending")
        self.assertEqual(res.content, response.content)

    def test_process_response_status_in_progress(self):
        task = self._create_task()
        api = self._create_instance(task)
        response = Response()
        response.status_code = 200
        response.raw = io.BytesIO(
            b'<?xml version="1.0" encoding="UTF-8"?><done>false</done>'
        )
        api.status = "InProgress"
        res = api._process_response_status(response)
        self.assertEqual(api.status, "InProgress")
        self.assertEqual(res.content, response.content)

    def test_process_response_status_in_progress_state_detail(self):
        task = self._create_task()
        api = self._create_instance(task)
        response = Response()
        response.status_code = 200
        response.raw = io.BytesIO(
            b'<?xml version="1.0" encoding="UTF-8"?><test><done>false</done><stateDetail>Deploy log goes here</stateDetail></test>'
        )
        api.status = "InProgress"
        res = api._process_response_status(response)
        self.assertEqual(api.status, "InProgress")
        self.assertEqual(res.content, response.content)


class TestBaseMetadataApiCall(BaseTestMetadataApi):
    def test_build_envelope_start_no_envelope(self):
        task = self._create_task()
        api = self._create_instance(task)
        with pytest.raises(AssertionError):
            api._build_envelope_start()

    def test_build_envelope_status_no_envelope(self):
        task = self._create_task()
        api = self._create_instance(task)
        self.assertEqual(api._build_envelope_status(), None)

    def test_build_envelope_result_no_envelope(self):
        task = self._create_task()
        api = self._create_instance(task)
        self.assertEqual(api._build_envelope_result(), None)

    def test_get_response_no_start_env(self):
        task = self._create_task()
        api = self._create_instance(task)
        with self.assertRaises(NotImplementedError):
            api._get_response()

    @responses.activate
    def test_get_response_no_status(self):
        org_config = {
            "instance_url": "https://na12.salesforce.com",
            "id": "https://login.salesforce.com/id/00D000000000000ABC/005000000000000ABC",
            "access_token": "0123456789",
        }
        task = self._create_task(org_config=org_config)
        api = self._create_instance(task)
        api.soap_envelope_start = "{api_version}"
        response = b'<?xml version="1.0" encoding="UTF-8"?><foo />'
        self._mock_call_mdapi(api, response)
        resp = api._get_response()
        self.assertEqual(resp.content, response)


class TestApiDeploy(BaseTestMetadataApi):
    api_class = ApiDeploy
    envelope_status = deploy_status_envelope

    def setUp(self):
        super(TestApiDeploy, self).setUp()
        self.package_zip = DummyPackageZipBuilder().as_base64()

    def _expected_envelope_start(self):
        return self.envelope_start.format(
            package_zip=self.package_zip,
            check_only="false",
            purge_on_delete="false",
            test_level="",
            run_tests="",
        )

    def _response_call_success_result(self, response_result):
        return deploy_result.format(status="Succeeded", extra="").encode()

    def _expected_call_success_result(self, response_result):
        return "Success"

    def _create_instance(
        self,
        task,
        api_version=None,
        purge_on_delete=None,
        check_only=None,
        test_level=None,
        run_tests=None,
    ):
        return self.api_class(
            task,
            self.package_zip,
            api_version=api_version,
            purge_on_delete=purge_on_delete,
            check_only=check_only,
            test_level=test_level,
            run_tests=run_tests,
        )

    def test_init_no_purge_on_delete(self):
        task = self._create_task()
        api = self._create_instance(task, purge_on_delete=False)
        self.assertEqual(api.purge_on_delete, "false")

    def test_init_default_check_only(self):
        task = self._create_task()
        api = self._create_instance(task)
        self.assertEqual(api.check_only, "false")

    def test_init_check_only(self):
        task = self._create_task()
        api = self._create_instance(task, check_only=True)
        self.assertEqual(api.check_only, "true")

    def test_init_default_test_level(self):
        task = self._create_task()
        api = self._create_instance(task)
        self.assertEqual(api.test_level, None)

    def test_init_test_level(self):
        task = self._create_task()
        api = self._create_instance(task, test_level="NoTestRun")
        self.assertEqual(api.test_level, "NoTestRun")

    def test_init_default_run_tests(self):
        task = self._create_task()
        api = self._create_instance(task)
        self.assertEqual(api.run_tests, [])

    def test_init_run_tests(self):
        task = self._create_task()
        api = self._create_instance(task, run_tests=["TestA", "TestB"])
        self.assertEqual(api.run_tests, ["TestA", "TestB"])

    def test_build_envelope_status__run_specified_tests(self):
        task = self._create_task()
        api = self._create_instance(
            task, run_tests=["TestA", "TestB"], test_level="RunSpecifiedTests"
        )

        api.package_zip = "Test"

        envelope = api._build_envelope_start()
        assert "<runTests>TestA</runTests>" in envelope
        assert "<runTests>TestB</runTests>" in envelope
        assert "RunSpecifiedTests" in envelope

    def test_process_response_metadata_failure(self):
        task = self._create_task()
        api = self._create_instance(task)
        response = Response()
        response.status_code = 200
        response.raw = io.BytesIO(
            deploy_result_failure.format(
                details="""<componentFailures>
  <problem>problem</problem>
  <problemType>Error</problemType>
  <componentType>CustomObject</componentType>
  <fileName>Test__c</fileName>
  <lineNumber>1</lineNumber>
  <columnNumber>1</columnNumber>
  <created>false</created>
  <deleted>false</deleted>
</componentFailures>"""
            ).encode()
        )
        with self.assertRaises(MetadataComponentFailure) as cm:
            api._process_response(response)
        expected = "Update of CustomObject Test__c: Error on line 1, col 1: problem"
        self.assertEqual(expected, str(cm.exception))

    def test_process_response_metadata_failure_no_lineno(self):
        task = self._create_task()
        api = self._create_instance(task)
        response = Response()
        response.status_code = 200
        response.raw = io.BytesIO(
            deploy_result_failure.format(
                details="""<componentFailures>
  <problem>problem</problem>
  <problemType>Error</problemType>
  <componentType>CustomObject</componentType>
  <fileName>Test__c</fileName>
  <created>false</created>
  <deleted>false</deleted>
</componentFailures>"""
            ).encode()
        )
        with self.assertRaises(MetadataComponentFailure) as cm:
            api._process_response(response)
        expected = "Update of CustomObject Test__c: Error: problem"
        self.assertEqual(expected, str(cm.exception))

    def test_process_response_metadata_failure_no_file_name(self):
        task = self._create_task()
        api = self._create_instance(task)
        response = Response()
        response.status_code = 200
        response.raw = io.BytesIO(
            deploy_result_failure.format(
                details="""<componentFailures>
  <problem>problem</problem>
  <problemType>Error</problemType>
  <componentType>CustomObject</componentType>
  <created>false</created>
  <deleted>false</deleted>
</componentFailures>"""
            ).encode()
        )
        with self.assertRaises(MetadataComponentFailure) as cm:
            api._process_response(response)
        expected = "Update of CustomObject: Error: problem"
        self.assertEqual(expected, str(cm.exception))

    def test_process_response_problem(self):
        task = self._create_task()
        api = self._create_instance(task)
        response = Response()
        response.status_code = 200
        response.raw = io.BytesIO(
            deploy_result_failure.format(
                details="""<problem>problem</problem>"""
            ).encode()
        )
        with self.assertRaises(MetadataApiError) as cm:
            api._process_response(response)
        expected = "problem"
        self.assertEqual(expected, str(cm.exception))

    def test_process_response_test_failure(self):
        task = self._create_task()
        api = self._create_instance(task)
        response = Response()
        response.status_code = 200
        response.raw = io.BytesIO(
            deploy_result_failure.format(
                details="""<runTestResult>
  <failures>
    <namespace>test</namespace>
    <stackTrace>stack</stackTrace>
  </failures>
</runTestResult>
"""
            ).encode()
        )
        with self.assertRaises(ApexTestException) as cm:
            api._process_response(response)
        expected = "Apex Test Failure: from namespace test: stack"
        self.assertEqual(expected, str(cm.exception))

    def test_process_response_no_status(self):
        task = self._create_task()
        api = self._create_instance(task)
        response = Response()
        response.status_code = 200
        response.raw = io.BytesIO(b"<bogus />")
        status = api._process_response(response)
        self.assertEqual(status, "Failed")

    def test_process_response_failure_but_no_message(self):
        task = self._create_task()
        api = self._create_instance(task)
        response = Response()
        response.status_code = 200
        response.raw = io.BytesIO(b"<status>Failed</status>")
        with self.assertRaises(MetadataApiError) as cm:
            api._process_response(response)
        self.assertEqual(response.text, str(cm.exception))

    def test_get_action(self):
        task = self._create_task()
        api = self._create_instance(task)
        self.assertEqual("Create", api._get_action(True, False))
        self.assertEqual("Delete", api._get_action(False, True))
        self.assertEqual("Update", api._get_action(False, False))


class TestApiListMetadata(BaseTestMetadataApi):
    api_class = ApiListMetadata
    envelope_start = list_metadata_start_envelope

    def setUp(self):
        super(TestApiListMetadata, self).setUp()
        self.metadata_type = "CustomObject"
        self.metadata = None
        self.folder = None
        self.api_version = self.project_config.project__package__api_version

    def _response_call_success_result(self, response_result):
        return list_metadata_result

    def _expected_call_success_result(self, response_result):
        metadata = defaultdict(list)
        metadata["CustomObject"] = [
            {
                u"createdById": None,
                u"createdByName": None,
                u"createdDate": datetime.datetime(2018, 8, 7, 16, 31, 57),
                u"fileName": None,
                u"fullName": u"Test__c",
                u"id": None,
                u"lastModifiedById": None,
                u"lastModifiedByName": None,
                u"lastModifiedDate": None,
                u"manageableState": None,
                u"namespacePrefix": None,
                u"type": "CustomObject",
            }
        ]
        return metadata

    def _create_instance(self, task, api_version=None):
        if api_version is None:
            api_version = self.api_version
        return self.api_class(
            task,
            metadata_type=self.metadata_type,
            metadata=self.metadata,
            folder=self.folder,
            as_of_version=api_version,
        )

    @responses.activate
    def test_bad_date_somehow(self):
        org_config = {
            "instance_url": "https://na12.salesforce.com",
            "id": "https://login.salesforce.com/id/00D000000000000ABC/005000000000000ABC",
            "access_token": "0123456789",
        }
        task = self._create_task(org_config=org_config)
        api = self._create_instance(task)

        self._mock_call_mdapi(api, list_metadata_result_bad_val)
        with self.assertRaises(MetadataParseError):
            api()


class TestApiRetrieveUnpackaged(BaseTestMetadataApi):
    maxDiff = None
    api_class = ApiRetrieveUnpackaged
    envelope_start = retrieve_unpackaged_start_envelope

    def setUp(self):
        super(TestApiRetrieveUnpackaged, self).setUp()
        self.package_xml = """<?xml version="1.0" encoding="UTF-8"?>
<Package xmlns="http://soap.sforce.com/2006/04/metadata">
    <version>41.0</version>
</Package>"""
        self.result_zip = DummyPackageZipBuilder()

    def _response_call_success_result(self, response_result):
        return retrieve_result.format(
            zip=self.result_zip.as_base64(), extra=""
        ).encode()

    def _expected_call_success_result(self, response_result):
        return self.result_zip.zf

    def _create_instance(self, task, api_version=None):
        return self.api_class(task, self.package_xml, api_version=api_version)

    @responses.activate
    def test_call_success(self):
        org_config = {
            "instance_url": "https://na12.salesforce.com",
            "id": "https://login.salesforce.com/id/00D000000000000ABC/005000000000000ABC",
            "access_token": "0123456789",
        }
        task = self._create_task(org_config=org_config)
        api = self._create_instance(task)
        if not self.api_class.soap_envelope_start:
            api.soap_envelope_start = "{api_version}"
        if not self.api_class.soap_envelope_status:
            api.soap_envelope_status = "{process_id}"
        if not self.api_class.soap_envelope_result:
            api.soap_envelope_result = "{process_id}"

        response = '<?xml version="1.0" encoding="UTF-8"?><id>1234567890</id>'
        self._mock_call_mdapi(api, response)
        response_status = '<?xml version="1.0" encoding="UTF-8"?><done>true</done>'
        self._mock_call_mdapi(api, response_status)
        response_result = '<?xml version="1.0" encoding="UTF-8"?><foo>bar</foo>'
        response_result = self._response_call_success_result(response_result)
        self._mock_call_mdapi(api, response_result)

        zip_file = api()

        self.assertEqual(
            zip_file.namelist(),
            self._expected_call_success_result(response_result).namelist(),
        )


class TestApiRetrieveInstalledPackages(BaseTestMetadataApi):
    api_class = ApiRetrieveInstalledPackages

    def _create_instance(self, task, api_version=None):
        api = self.api_class(task, api_version)
        return api

    def _expected_call_success_result(self, result_response):
        return {}

    def test_process_response_no_zipstr(self):
        task = self._create_task()
        api = self._create_instance(task)
        response = Response()
        response.status_code = 200
        response.raw = io.BytesIO(
            deploy_result.format(status="testing", extra="").encode()
        )
        resp = api._process_response(response)
        self.assertEqual(resp, {})

    def test_process_response_zipstr_no_packages(self):
        task = self._create_task()
        api = self._create_instance(task)
        response = Response()
        response.status_code = 200
        response.raw = io.BytesIO(
            retrieve_result.format(
                zip=CreatePackageZipBuilder("testing", api.api_version).as_base64(),
                extra="",
            ).encode()
        )
        resp = api._process_response(response)
        self.assertEqual(resp, {})

    def test_process_response_zipstr_one_package(self):
        task = self._create_task()
        api = self._create_instance(task)
        response = Response()
        response.status_code = 200
        response.raw = io.BytesIO(
            retrieve_result.format(
                zip=InstallPackageZipBuilder("foo", "1.1").as_base64(), extra=""
            ).encode()
        )
        resp = api._process_response(response)
        self.assertEqual(resp, {"foo": "1.1"})


class TestApiRetrievePackaged(TestApiRetrieveUnpackaged):
    api_class = ApiRetrievePackaged
    envelope_start = retrieve_packaged_start_envelope

    def setUp(self):
        super(TestApiRetrievePackaged, self).setUp()
        self.package_name = "Test Package"

    def _expected_envelope_start(self):
        return self.envelope_start.format(
            api_version=self.project_config.project__package__api_version,
            package_name=self.package_name,
        )

    def _create_instance(self, task, api_version=None):
        return self.api_class(task, self.package_name, api_version)
